#!/usr/bin/env python
# -*- coding: utf-8 -*-

import itertools
import functools
import types
import numpy as np


def mod(a, b):
    """
    Returns a % b, with a % 0 = a.

    Parameters
    ----------
    a : int
        The first argument in a % b.
    b : int
        The second argument in a % b.

    Returns
    -------
    int
        `a` modulus `b`.

    Examples
    ---------
    >>> mod(5, 2)
    1
    >>> mod(5, 0)
    5
    """
    if b == 0:
        return a
    return a % b


def verify_dims_list(list_of_lists, dims):
    """
    Verify the dimensions of a list of lists.

    Parameters
    ----------
    list_of_lists : list
        A nested list of lists.
    dims : list
        A list of dimensions.

    Returns
    -------
    verified : bool
        Whether or not the dimensions match the `dims` parameter.

    Examples
    ---------
    >>> table = [1 ,2 ,3]
    >>> dims = [3]
    >>> verify_dims_list(table, dims)
    True

    >>> table = [[1, 2], [1, 2], [1, 2]]
    >>> dims = [3, 2]
    >>> verify_dims_list(table, dims)
    True

    >>> table = [[[1, 2, 3, 4], [1, 2, 3, 4]],
    ...          [[1, 2, 3, 4], [1, 2, 4]],
    ...          [[1, 2, 3, 4], [1, 2, 3, 4]]]
    >>> dims = [3, 2, 4] # Not correct, notice the missing value above
    >>> verify_dims_list(table, dims)
    False
    """
    for i, dim in enumerate(dims):
        product_arg = [range(dims[k]) for k in range(i)]
        for p in itertools.product(*product_arg):
            if not len(call_nested_list(list_of_lists, p)) == dim:
                return False

    return True


def call_nested_list(list_of_lists, arg):
    """
    Call a nested list like a function.

    Parameters
    ----------
    list_of_lists : list
        A nested list of lists.
    arg : list
        The argument [dim1, dim2, ...].

    Returns
    -------
    value : object
        The object in the list of lists.

    Examples
    ---------
    >>> table = [1 ,2 ,3]
    >>> call_nested_list(table, [0])
    1

    >>> table = [[1, 2], [1, 2], [1, 2]]
    >>> call_nested_list(table, [0, 0])
    1

    >>> table = [[[1, 2, 3, 4], [1, 2, 3, 4]],
    ...          [[1, 2, 3, 4], [1, 2, 3, 4]],
    ...          [[1, 2, 3, 4], [1, 2, 3, 4]]]
    >>> call_nested_list(table, [0, 0])
    [1, 2, 3, 4]
    >>> call_nested_list(table, [0, 0, 0])
    1
    """

    answer = list_of_lists
    for index in arg:
        answer = answer[index]
    return answer


def function_to_table(function, dims, *args, **kwargs):
    """

    Parameters
    ----------
    function : function
        A function with signature `(list_arg, *args, **kwargs)`.
    dims : list
        A list of dimensions such as [8, 6, 3].

    Returns
    -------

    """

    # Create an empty table using numpy
    table = np.empty(dims, dtype=complex)

    # Iterate through the domain and evaluate the function
    for list_arg in itertools.product(*[range(d) for d in dims]):
        table[list_arg] = function(list(list_arg),  *args, **kwargs)

    return table


def arg(min_or_max, iterable, function_of_element):
    """
    Call a nested list like a function.

    Parameters
    ----------
    list_of_lists : list
        A nested list of lists.
    arg : list
        The argument [dim1, dim2, ...].

    Returns
    -------
    value : object
        The object in the list of lists.

    Examples
    ---------
    >>> iterable = [-8, -4, -2, 3, 5]
    >>> arg(min, iterable, abs)
    -2
    >>> iterable = range(-10, 10)
    >>> arg(max, iterable, lambda x: -(x - 3)**2)
    3
    """
    return min_or_max(iterable, key=function_of_element)


def copy_func(f):
    """Based on http://stackoverflow.com/a/6528148/190597 (Glenn Maynard)"""
    g = types.FunctionType(f.__code__, f.__globals__, name=f.__name__,
                           argdefs=f.__defaults__,
                           closure=f.__closure__)
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    return g


argmin = functools.partial(arg, min_or_max = min)
argmax = functools.partial(arg, min_or_max = max)

if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose = True)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Dummy conftest.py for abelian.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    https://pytest.org/latest/plugins.html
"""
from __future__ import print_function, absolute_import, division

import pytest


def pytest_addoption(parser):
    parser.addoption('--repeat', action='store',
        help='Number of times to repeat each test')


def pytest_generate_tests(metafunc):
    if metafunc.config.option.repeat is not None:
        count = int(metafunc.config.option.repeat)

        # We're going to duplicate these tests by parametrizing them,
        # which requires that each test has a fixture to accept the parameter.
        # We can add a new fixture like so:
        metafunc.fixturenames.append('tmp_ct')

        # Now we parametrize. This is what happens when we do e.g.,
        # @pytest.mark.parametrize('tmp_ct', range(count))
        # def test_foo(): pass
        metafunc.parametrize('tmp_ct', range(count))


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module consists of a class for functions on LCAs,
called LCAFunc. Such a function represents a function
from a LCA G to the complex numbers C.
"""
from operator import itemgetter
from sympy import Matrix, Float, Integer, Add, Rational
from abelian.linalg import solvers, free_to_free
from abelian.linalg.utils import norm, difference
from abelian.linalg.free_to_free import elements_increasing_norm
from abelian.linalg.solvers import solve
from abelian.utils import call_nested_list, verify_dims_list, copy_func, function_to_table
from abelian.groups import LCA
from types import FunctionType
from collections.abc import Callable
import numpy as np
import functools
import operator


class LCAFunc(Callable):
    """
    A function from an LCA to a complex number.
    """

    def __init__(self, representation, domain):
        """
        Initialize a function G -> C.

        Parameters
        ----------
        representation : function or n-dimensional list of domain allows it
            A function which takes in a list as a first argument, representing
            the group element. Alternatively a list of lists if the domain is
            discrete and of finite order.
        domain : LCA
            An elementary locally compact abelian group, which is the domain
            of the function.

        Examples
        ---------

        If a function representation is used, functions on domains are
        relatively straightforward.

        >>> def power(list_arg, exponent = 2):
        ...     return sum(x**exponent for x in list_arg)
        >>> from abelian import LCAFunc, LCA
        >>> # A function on R/Z = T
        >>> f = LCAFunc(power, LCA([1], [False]))
        >>> f([0.5])
        0.25
        >>> f([1.5], exponent = 3) == 0.5**3
        True
        >>> # A function on Z_p
        >>> f = LCAFunc(power, LCA([5, 10]))
        >>> f([1,1]) == f([6, 11])
        True
        >>> f([2, 2], exponent = 1)
        4

        If a table representation is used, the function can be defined on
        direct sums of Z_n.

        >>> # Define a table: a list of lists
        >>> table = [[1, 2],
        ...          [3, 4],
        ...          [5, 6]]
        >>> f = LCAFunc(table, LCA([3, 2]))
        >>> f([1, 1])
        4
        >>> f([3, 1])
        2
        >>> import numpy as np
        >>> f = LCAFunc(np.array(table), LCA([3, 2]))
        >>> f([1, 1])
        4
        """

        # Verify that the domain is an LCA
        if not isinstance(domain, LCA):
            raise TypeError('Domain must be an LCA instance.')

        self.domain = domain

        # A function representation has been passed
        if isinstance(representation, FunctionType):
            # A function representation has been passed
            self.representation = representation
            self.table = None

        # A table representation has been passed
        else:
            if not self._discrete_finite_domain():
                raise TypeError('When the function representation is a table,'
                                'the domain must be a FGA with finite '
                                'orders.')

            # Verify the dimensions of the data table
            if not verify_dims_list(representation, self.domain.orders):
                raise ValueError('Table dimension mismatch.')

            # Return a callable data table
            def list_caller(list_of_points):
                return call_nested_list(representation, list_of_points)

            self.representation = list_caller
            self.table = representation


    def __call__(self, list_arg, *args, **kwargs):
        """
        Override function calls,
        see :py:meth:`~abelian.functions.LCAFunc.evaluate`.
        """
        return self.evaluate(list_arg, *args, **kwargs)

    def __repr__(self):
        """
        Override the ``repr()`` function.

        Returns
        -------
        representation :str
            A representation of the instance.

        """
        str = r'LCAFunc on domain {}'.format(self.domain)
        return str

    def copy(self):
        """
        Return a copy of the instance.

        Returns
        -------
        function : LCAFunc
            A copy of `self`.

        Examples
        --------
        >>> from abelian import LCA, LCAFunc
        >>> f = LCAFunc(lambda x:sum(x), LCA([0]))
        >>> g = f.copy()
        >>> f([1]) == g([1])
        True
        """
        repr = copy_func(self.representation)
        domain = self.domain.copy()
        return type(self)(representation = repr, domain = domain)

    def dft(self, func_type = None):
        """
        If the domain allows it, compute DFT.

        This method uses the n-dimensional Fast Fourier Transform (FFT) to
        compute the n-dimensional Discrete Fourier Transform. The data is
        converted to a :py:class:`~numpy.ndarray` object for efficient
        numerical computation, then the :py:func:`~numpy.fft.fftn` function
        is used to compute the fast fourier transform.

        This implementation is different from the implementation in
        :py:func:`~numpy.fft.fftn` by a factor. While the :py:func:`~numpy.fft.fftn`
        function divides by m*n on the inverse transform, this implementation
        does it on the forward transform, and vice verca.


        Parameters
        ----------
        func_type : str
            If None, compute the function values using pure python.
            If 'ogrid', use a numpy.ogrid (open mesh-grid) to compute the
            functino values.
            If 'mgrid', use a numpy.mgrid (dense mesh-grid) to compute the
            function values.

        Returns
        -------
        function : LCAFunc
            The discrete Fourier transformation of the original function.


        Examples
        --------
        >>> from abelian import LCA, LCAFunc
        >>> # Create a simple linear function on Z_5 + Z_4 + Z_3
        >>> domain = LCA([5, 4, 3])
        >>> def linear(list_arg):
        ...     return sum(list_arg)
        >>> func = LCAFunc(linear, domain)
        >>> func([1, 2, 1])
        4
        >>> # Take the discrete fourier transform and evaluate
        >>> func_dft = func.dft()
        >>> func_dft([0, 0, 0])
        (4.5+0j)
        >>> # Take the inverse discrete fourier transform
        >>> func_dft_idft = func_dft.idft()
        >>> # Numerics might not make this equal, but mathematically it is
        >>> abs(func_dft_idft([1, 2, 1]) - func([1, 2, 1])) < 10e-10
        True
        """
        return self._fft_wrapper(func_to_wrap='fftn', func_type=func_type)



    def evaluate(self, list_arg, *args, **kwargs):
        """
        Evaluate function on a group element.

        Parameters
        ----------
        list_arg : list
            The first argument, which must be a list (interpreted as vector).
        *args : tuple
            An unpacked tuple of arguments.
        **kwargs : dict
            An unpacked dictionary of arguments.

        Returns
        -------
        value : complex, float or int
            A complex number (could be real or integer).

        Examples
        --------
        >>> from abelian import LCA, LCAFunc
        >>> R = LCA([0], [False])
        >>> function = LCAFunc(lambda x: 1, domain = R**2)
        >>> function([1, 2])
        1

        Some subtle concepts are shown below.

        >>> function(1)
        Traceback (most recent call last):
        ...
        ValueError: Argument to function must be list.
        >>> function([1])
        Traceback (most recent call last):
        ...
        ValueError: LCAFunc argument does not match domain length.
        >>> type(function([1, 1])) in (int, float, complex)
        True
        """

        # If the domain consists of more than one group in the direct sum,
        # the argument must have the same length. If the direct sum consists
        # of one group only and the argument is numeric, we forgive the user
        domain_length = self.domain.length()

        # Verify the inputs
        if domain_length > 1:
            if isinstance(list_arg, (int, float, complex)):
                raise ValueError('Argument to function must be list.')
        elif domain_length and isinstance(list_arg, (int, float, complex)):
            list_arg = [list_arg]

        # Verify the length
        if domain_length != len(list_arg):
            raise ValueError('LCAFunc argument does not match domain length.')

        # Project and compute
        proj_args = self.domain.project_element(list_arg)
        answer = self.representation(proj_args, *args, **kwargs)

        # Cast to Python data type
        if isinstance(answer, (Float, Rational)):
            return float(answer)
        if isinstance(answer, Integer):
            return int(answer)
        if isinstance(answer, Add):
            return complex(answer)
        return answer

    def idft(self, func_type = None):
        """
        If the domain allows it, compute inv DFT.

        This is a wrapper around np.fft.ifftn.

        Parameters
        ----------
        func_type : str
            If None, compute the function values using pure python.
            If 'ogrid', use a numpy.ogrid (open mesh-grid) to compute the
            functino values.
            If 'mgrid', use a numpy.mgrid (dense mesh-grid) to compute the
            function values.

        Returns
        -------
        function : LCAFunc
            The inverse discrete Fourier transformation of the original
            function.


        Examples
        --------
        >>> from abelian import LCA, LCAFunc
        >>> # Create a simple linear function on Z_5 + Z_4 + Z_3
        >>> domain = LCA([5, 4, 3])
        >>> def linear(list_arg):
        ...     x, y, z = list_arg
        ...     return complex(x + y, z - x)
        >>> func = LCAFunc(linear, domain)
        >>> func([1, 2, 1])
        (3+0j)
        >>> func_idft = func.idft()
        >>> func_idft([0, 0, 0])
        (210-60j)
        """
        return self._fft_wrapper(func_to_wrap='ifftn', func_type=func_type)


    def pointwise(self, other, operator):
        """
        Apply pointwise binary operator.

        Parameters
        ----------
        other : LCAFunc
            Another Functin on the same domain.
        operator : function
            A binary operator.

        Returns
        -------
        function : LCAFunc
            The resulting function, new = operator(self, other).

        Examples
        --------
        >>> from abelian import LCA
        >>> domain = LCA([5])
        >>> function1 = LCAFunc(lambda arg: sum(arg), domain)
        >>> function2 = LCAFunc(lambda arg: sum(arg)*2, domain)
        >>> from operator import add
        >>> pointwise_add = function1.pointwise(function2, add)
        >>> function1([2]) + function2([2]) == pointwise_add([2])
        True
        >>> from operator import mul
        >>> sample_points = [0, 1, 2, 3, 4]
        >>> pointwise_mul = function1.pointwise(function2, mul)
        >>> pointwise_mul.sample(sample_points) # i * 2*i = 2*i*i
        [0, 2, 8, 18, 32]

        """
        if self.domain != other.domain:
            raise ValueError('Domains must be equal.')

        # Perform both function and apply the operator
        def new_repr(list_arg, *args, **kwargs):
            result_self = self.representation(list_arg, *args, **kwargs)
            result_other = other.representation(list_arg, *args, **kwargs)
            return operator(result_self, result_other)

        return type(self)(domain = self.domain, representation = new_repr)

    def pullback(self, morphism):
        """
        Return the pullback along `morphism`.

        The pullback is the composition `morphism`, then `self`.
        The domain of `self` must match the target of the morphism.

        Parameters
        ----------
        morphism : HomLCA
            A homomorphism between LCAs

        Returns
        -------
        pullback : LCAFunc
            The pullback of `self` along `morphism`.

        Examples
        --------
        Using a simple function and homomorphism.

        >>> from abelian import HomLCA, LCA
        >>> # Create a function on Z
        >>> f = LCAFunc(lambda list_arg:list_arg[0]**2, LCA([0]))
        >>> # Create a homomorphism from Z to Z
        >>> phi = HomLCA([2])
        >>> # Pull f back along phi
        >>> f_pullback = f.pullback(phi)
        >>> f_pullback([4]) == 64 # (2*4)**2 == 64
        True

        Using a simple function and homomorphism represented as matrix.

        >>> from abelian import HomLCA, LCA
        >>> def func(list_arg):
        ...     x, y = tuple(list_arg)
        ...     return x ** 2 + y ** 2
        >>> domain = LCA([5, 3])
        >>> f = LCAFunc(func, domain)
        >>> phi = HomLCA([1, 1], target=domain)
        >>> f_pullback = f.pullback(phi)
        >>>
        >>> f_pullback([8]) == 13
        True
        """

        if not self.domain == morphism.target:
            raise ValueError('Target of morphism must equal domain of '
                             'function.')

        # Get the domain for the new function
        domain = morphism.source

        def new_repr(list_arg, *args, **kwargs):
            """
            A function which first applies the morphism,
            then applies the function.
            """
            applied_morph = morphism.evaluate(list_arg)
            applied_func = self.representation(applied_morph, *args, **kwargs)
            return applied_func

        return type(self)(representation = new_repr, domain = domain)

    def pushforward(self, morphism, terms_in_sum = 50):
        """
        Return the pushforward along `morphism`.

        The pushforward is computed by solving an equation, finding the
        kernel, and iterating through the kernel. The pushfoward
        approximates a possibly infinite sum by `terms_in_sum` terms.


        Parameters
        ----------
        morphism : HomLCA
            A homomorphism between LCAs.
        terms_in_sum : int
            The number of terms in the sum to use, i.e. the number of solutions
            to the equation to iterate over.
        norm_condition : function
            If not None, a function can be used to terminate the sum.
            The norm_condition must be a function of a group element,
            and when the function is false for every v in the kernel
            such that maxnorm(v) = C for a given C, then the sum terminates.


        Returns
        -------
        pushforward : LCAFunc
            The pushforward of `self` along `morphism`.

        Examples
        --------

        The first example is a homomorphism R -> T.

        >>> from abelian import LCA, LCAFunc, HomLCA
        >>> R = LCA([0], [False])
        >>> T = LCA([1], [False])
        >>> epimorphism = HomLCA([1], source = R, target = T)
        >>> func_expr = lambda x: 2**-sum(x_j**2 for x_j in x)
        >>> func = LCAFunc(func_expr, domain = R)
        >>> func.pushforward(epimorphism, 1)([0]) # 1 term in the sum
        1.0
        >>> func.pushforward(epimorphism, 3)([0]) # 1 + 0.5*2
        2.0
        >>> func.pushforward(epimorphism, 5)([0]) # 1 + 0.5*2 + 0.0625*2
        2.125

        The first example is a homomorphism Z -> Z_2.

        >>> from abelian import LCA, LCAFunc, HomLCA
        >>> Z = LCA([0], [True])
        >>> Z_2 = LCA([2], [True])
        >>> epimorphism = HomLCA([1], source = Z, target = Z_2)
        >>> func_expr = lambda x: 2**-sum(x_j**2 for x_j in x)
        >>> func = LCAFunc(func_expr, domain = Z)
        >>> func.pushforward(epimorphism, 1)([0]) # 1 term in the sum
        1.0
        >>> func.pushforward(epimorphism, 3)([0]) # 1 + 0.5*2 + 0.0625*2
        1.125

        The third example is a homomorphism R -> R.

        >>> from abelian import LCA, LCAFunc, HomLCA
        >>> R = LCA([0], [False])
        >>> epimorphism = HomLCA([1], source = R, target = R)
        >>> func_expr = lambda x: 2**-sum(x_j**2 for x_j in x)
        >>> func = LCAFunc(func_expr, domain = R)
        >>> func.pushforward(epimorphism, 3)([0]) # 1 term in the sum
        1.0


        """
        if not self.domain == morphism.source:
            raise ValueError('Source of morphism must equal domain of '
                             'function.')

        # Compute the kernel
        kernel = morphism.kernel()
        kernel_m, kernel_n = kernel.A.shape

        # Get the domain for the new function
        domain = morphism.target

        def new_representation(list_arg, *args, **kwargs):
            """
            A function which first applies the morphism,
            then applies the function.
            """
            # Compute a solution to phi(x) = y
            target_orders = Matrix(morphism.target.orders)
            base_ans = solve(morphism.A, Matrix(list_arg), target_orders)

            # The kernel is empty, do not start summing, just return
            if kernel_n == 0:
                return self.representation(base_ans, *args, **kwargs)

            # Iterate through the kernel space and compute the sum
            kernel_sum = 0
            dim_ker_source = len(kernel.source)

            generator = elements_increasing_norm(dim_ker_source)
            for counter, boundary_element in enumerate(generator, 1):
                # The `base_ans` is in the kernel of the morphism,
                # we move to all points in the kernel by taking
                # the `base_ans` + linear combinations of the kernel
                linear_comb = Matrix(list(boundary_element))
                kernel_element = base_ans + kernel.evaluate(linear_comb)

                function = self.representation
                func_in_ker = function(kernel_element, *args, **kwargs)
                kernel_sum += func_in_ker
                if counter >= terms_in_sum:
                    break

            return kernel_sum

        return type(self)(representation=new_representation, domain=domain)


    def sample(self, list_of_elements, *args, **kwargs):
        """
        Sample on a list of group elements.

        Parameters
        ----------
        list_of_elements : list
            A list of groups elements, where each element is also a list.

        Returns
        -------
        sampled_vals : list
            A list of sampled values at the elements.

        Examples
        --------
        >>> from abelian import LCAFunc, LCA
        >>> func = LCAFunc(lambda x : sum(x), LCA([0, 0]))
        >>> sample_points = [[0, 0], [1, 2], [2, 1], [3, 3]]
        >>> func.sample(sample_points)
        [0, 3, 3, 6]
        """

        return [self.evaluate(p, *args, **kwargs) for p in list_of_elements]


    def shift(self, list_shift):
        """
        Shift the function.


        Parameters
        ----------
        list_shift : list
            A list of shifts.

        Returns
        -------
        function : LCAFunc
            A new function which is shifted.

        Examples
        --------
        >>> from abelian import LCAFunc, LCA
        >>> func = LCAFunc(lambda x: sum(x), LCA([0]))
        >>> func.sample([0, 1, 2, 3])
        [0, 1, 2, 3]
        >>> func.shift([2]).sample([0, 1, 2, 3])
        [-2, -1, 0, 1]
        """
        new_domain = self.domain

        def new_representation(list_arg, *args, **kwargs):
            """
            A function which first applies the morphism,
            then applies the function.
            """
            generator = zip(list_arg, list_shift)
            shifted_arg = [arg-shift for (arg, shift) in generator]
            applied_func = self.representation(shifted_arg, *args, **kwargs)
            return applied_func


        return type(self)(representation = new_representation,
                          domain = new_domain)


    def to_latex(self):
        """
        Return as a :math:`\LaTeX` string.


        Returns
        -------
        latex_str : str
            The object as a latex string.

        """
        latex_str = r'\operatorname{function} \in \mathbb{C}^G, \ G = GRP'
        latex_str = latex_str.replace('GRP', self.domain.to_latex())
        return latex_str

    def to_table(self, *args, **kwargs):
        """
        Return a n-dimensional table.

        Returns
        -------
        table : n-dimensional list
            The table representation.

        Examples
        --------
        >>> from abelian import LCA, LCAFunc
        >>> domain = LCA([5, 5])
        >>> f = LCAFunc(lambda x: sum(x), domain)
        >>> table = f.to_table()
        >>> table[1][1]
        (2+0j)

        Using a table from the start.

        >>> from abelian import LCA, LCAFunc
        >>> import numpy as np
        >>> domain = LCA([5, 5])
        >>> f = LCAFunc(np.eye(5), domain)
        >>> table = f.to_table()
        >>> table[1][1]
        1.0
        >>> type(table)
        <class 'numpy.ndarray'>
        >>> f = LCAFunc([[1, 2], [2, 4]], LCA([2, 2]))
        >>> f.to_table()
        [[1, 2], [2, 4]]
        """

        # If the domain is not discrete and of finite order, no table exists
        if not self._discrete_finite_domain():
            raise TypeError('No table. Domain must be discrete and finite.')

        # If a table already is computed, return it
        if self.table is not None:
            return self.table

        # If a table is not computed, compute it and return
        dims = self.domain.orders
        table = function_to_table(self.representation, dims, *args, **kwargs)
        self.table = table
        return table

    def transversal(self, epimorphism, transversal_rule = None,
                    default_value = 0):
        """
        Pushforward using transversal rule.

        If (transversal * epimorphism)(x) = x, then x is pushed forward
        using the transversal rule. If not, then the default_value value is
        returned.

        Parameters
        ----------
        epimorphism : HomLCA
            An epimorphism.
        transversal_rule : function
            A function with signature `func(list_arg, *args, **kwargs)`.

        Returns
        -------
        function : LCAFunc
            The pushforward of `self` along the transversal of the epimorphism.

        Examples
        --------
        >>> from abelian import LCA, LCAFunc, HomLCA
        >>> n = 5 # Sice of the domain, Z_n
        >>> f_on_Zn = LCAFunc(lambda x: sum(x)**2, LCA([n]))
        >>> # To move this function to Z, create an epimorphism and a
        >>> # transversal rule
        >>> epimorphism = HomLCA([1], source = [0], target = [n])
        >>> def transversal_rule(x):
        ...     if sum(x) < n/2:
        ...         return [sum(x)]
        ...     elif sum(x) >= n/2:
        ...         return [sum(x) - n]
        ...     else:
        ...         return None
        >>> # Do the pushforward with the transversal rule
        >>> f_on_Z = f_on_Zn.transversal(epimorphism, transversal_rule)
        >>> f_on_Z.sample(list(range(-n, n+1)))
        [0, 0, 0, 9.0, 16.0, 0.0, 1.0, 4.0, 0, 0, 0]

        """
        new_domain = epimorphism.source

        if not transversal_rule:
            transversal_rule = voronoi(epimorphism, norm_p = 2)

        def new_representation(list_arg, *args, **kwargs):
            # Compose (section * transversal)(x)
            applied_epi = epimorphism.evaluate(list_arg)
            composed = transversal_rule(applied_epi)

            # If the composition is the identity, apply the epimorphism
            # and then the function to evaluate the new function at the point
            epsilon = 10e-10
            if difference(composed,list_arg) < epsilon:#composed == list_arg:
                return self.representation(applied_epi, *args, **kwargs)
            else:
                return default_value

        return type(self)(representation=new_representation, domain=new_domain)

    def _discrete_finite_domain(self):
        """
        Whether or not the domain is discrete and of finite order.

        Returns
        -------
        discrete_finite : bool
            Whether or not the domain is discrete and finite.
        """
        return self.domain.is_FGA() and all(p > 0 for p in self.domain.orders)


    def _fft_wrapper(self, func_to_wrap = 'fftn', func_type = ''):
        """
        Common wrapper for FFT and IFFT routines.

        The numpy DFT is defined as:
        :math:`A_{kl} =  \sum_{m=0}^{M-1} \sum_{n=0}^{N-1}
        a_{mn}\exp\left\{-2\pi i \left({mk\over M}+{nl\over N}\right)\right\}
        \qquad k = 0, \ldots, M-1;\quad l = 0, \ldots, N-1.`

        And the inverse DFT is defined as:
        :math:`a_{mn} = \frac{1}{MN} \sum_{k=0}^{M-1} \sum_{l=0}^{N-1}
        A_{kl}\exp\left\{2\pi i \left({mk\over M}+{nl\over N}\right)\right\}
        \qquad m = 0, \ldots, M-1;\quad n = 0, \ldots, N-1.`


        Parameters
        ----------
        func_to_wrap : str
            Name of the function from the np.fft library to call.
        func_type : str
            If empty, compute the function values using pure python.
            If 'ogrid', use a numpy.ogrid (open mesh-grid) to compute the
            functino values.
            If 'mgrid', use a numpy.mgrid (dense mesh-grid) to compute the
            function values.

        Returns
        -------
        function : LCAFunc
            The function with a numpy routine applied to every element
            in the domain.

        """

        # Verify that the inputs are sensible
        domain = self.domain
        dims = domain.orders
        if not all(p > 0 for p in dims) and domain.is_FGA():
            return ValueError('Domain must be discrete and of finite order.')

        # Put the function values in a table in preparation for FFT/IFFT
        if func_type is None:
            table = self.to_table()
        else:
            # Here the np.ogrid or np.mgrid can be used, see
            # https://arxiv.org/pdf/1102.1523.pdf
            function = getattr(np, func_type)
            table = function([tuple([slice(k) for k in dims])])

        # Take fft and convert to list of lists
        function_wrapped = getattr(np.fft, func_to_wrap, None)
        if function_wrapped is None:
            raise ValueError('Could not wrap:', func_to_wrap)
        table_computed = function_wrapped(table)

        # Scale differently then the Numpy implementation
        # Numpy divides by prod(dims) when computing the inverse,
        # but we do it when we compute the forward transform
        if func_to_wrap == 'fftn':
            table_computed =  table_computed / (functools.reduce(
                operator.mul, dims))
        elif func_to_wrap == 'ifftn':
            table_computed =  table_computed * (functools.reduce(
                operator.mul, dims))

        # Create a new instance and return
        return type(self)(domain = domain, representation = table_computed)


def voronoi(epimorphism, norm_p=2):
    """
    Return the Voronoi transversal function.

    This higher-order function returns a quotient transversal
    which maps x to the y which is cloest to the low-frequency
    fourier mode.

    Parameters
    ----------
    epimorphism_kernel : HomLCA
        The kernel of the epimorphism that we want to find a section for.
    norm : function or None
        A norm function, if None, the max-norm is used.

    Returns
    -------
    sigma : function
        A function x -> y.

    Examples
    ---------
    >>> # An orthogonal example
    >>> from abelian import LCAFunc, HomLCA, LCA
    >>> Z_10 = LCA([10])
    >>> epimorphism = HomLCA([1], target = Z_10)
    """

    # Verify that the epimorphism is R^n -> T^n or Z^n -> Z_n
    if not all(o == 0 for o in epimorphism.source.orders):
        raise ValueError('Epimorphism must be R^n -> T^n or Z^n -> Z_n')
    if not all(o >= 1 for o in epimorphism.target.orders):
        raise ValueError('Epimorphism must be R^n -> T^n or Z^n -> Z_n')

    # Solve for the kernel of the epimorphism
    epimorphism_kernel = epimorphism.kernel()

    def sigma(x):
        """
        A section used for quotient transversal.
        """
        # Initialize variables
        x = Matrix(x)
        kernel_A = epimorphism_kernel.A
        m, m = kernel_A.shape

        # STEP 1: Solve the equation phi(y) = x for the y
        # If the epimorphism is Z^n -> Z_n
        if epimorphism._is_homFGA():
            # Use the solver for FGAs
            p = Matrix(epimorphism.target.orders)
            y = solvers.solve(epimorphism.A, x, p = p)

        # If R^n -> T^n
        if (all(d == False for d in epimorphism.source.discrete) and
            all(d == False for d in epimorphism.target.discrete)):
            # Normal linear algebra, user sympy solver
            y = epimorphism.A.solve(x)

        # STEP 2: Find the points with max-norm <= 1
        points = free_to_free.elements_increasing_norm(m, end_value = 2)
        points = (Matrix(p) for p in points)

        # STEP 3: Find the corner that minimizes ||y - ker(epi) * points||
        funcvals_vals = ((y - kernel_A*p, p) for p in points)

        normvals_vals = ((norm(list(val), norm_p), p) for (val, p) in funcvals_vals)
        min_value, minimizer = min(normvals_vals, key=itemgetter(0))

        y = y - kernel_A*minimizer

        return list(y)

    return sigma


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose = False)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from random import randint as ri
import random
from sympy import Matrix
from abelian.linalg.utils import vector_mod_vector, remove_cols
from abelian.linalg.solvers import solve, solve_epi
from abelian.linalg.factorizations import hermite_normal_form


class TestSolveEpi():
    """
    Test the solver for the unknown epimorphism.
    This is used for the coimage.
    """

    @classmethod
    def setup_class(cls):
        # Set up the sizes for the tests
        cls.m = 3
        cls.k = 2


    def test_bijective_free_to_free(self):
        """
        Solve X*A = B when A is a bijective, free-to-free morphism.
        """
        m, k = self.m, self.k
        A = Matrix(m, m, lambda i, j : ri(-9, 9))
        A, H = hermite_normal_form(A)
        X = Matrix(k, m, lambda i, j : ri(-9, 9))
        B = X * A

        X_sol = solve_epi(A, B)
        assert B == X_sol * A


    def test_epi_free_to_free(self):
        """
        Solve X*A = B when A is overdetermined, free-to-free morphism.
        """
        m, k = self.m, self.k

        # Create a matrix A
        A = Matrix(m, m, lambda i, j : ri(-9, 9))
        A, H = hermite_normal_form(A)
        extra_cols = ri(1, 3)

        # Add extra columns, linear combination of existing
        lin_combs = Matrix(m, extra_cols, lambda i, j: ri(-9, 9))
        A = A * (Matrix.eye(m).row_join(lin_combs))
        X = Matrix(k, m, lambda i, j : ri(-9, 9))
        B = X * A

        X_sol = solve_epi(A, B)
        assert B == X_sol * A


class TestSolve():
    """
    Test the general solver.
    """

    @classmethod
    def setup_class(cls):
        # Set up the sizes for the tests
        cls.n = ri(3, 5)
        cls.r = cls.n - 2


    def test_bijective(self):
        """
        Test the equation solver when A is n x n and of full rank.
        """

        # Set up the equation
        n = self.n
        A = Matrix(n, n, lambda i,j : ri(-9,9))
        A, H = hermite_normal_form(A)
        x = Matrix(n, 1, lambda i, j : ri(-5,5))
        p = Matrix(n, 1, lambda i, j : ri(10,100))
        b = vector_mod_vector(A*x, p)

        # Solve the equation and verify
        x_sol = solve(A, b, p)
        assert (vector_mod_vector(A*x_sol, p) == b)

    def test_surjective(self):
        """
        Test the equation solver when A is surjective,
        i.e. maps onto (epimorphism).
        """

        # Create a matrix A
        n = self.n
        A = Matrix(n, n, lambda i, j: ri(-9, 9))
        A, H = hermite_normal_form(A)
        m, m = A.shape
        extra_cols = ri(1, 3)

        # Add extra columns, linear combinations of existing columns
        A = A * (Matrix.eye(m).row_join(Matrix(m, extra_cols, lambda i,j :ri(-9, 9))))
        p = Matrix(m, 1, lambda i, j: ri(10, 100))
        x = Matrix(m + extra_cols, 1, lambda i, j: ri(-5, 5))
        b = vector_mod_vector(A * x, p)

        x_sol = solve(A, b, p)
        assert (vector_mod_vector(A * x_sol, p) == b)

    def test_injective(self):
        """
        Test the equation solver when a is surjective,
        i.e. maps one-to-one (monomorphism).
        """

        # Create a unimodular matrix A
        n = self.n
        A = Matrix(n, n, lambda i, j: ri(-9, 9))
        A, H = hermite_normal_form(A)
        m, m = A.shape
        num_cols_remove = ri(1, m-1)

        # Remove columns from A
        cols_to_del = random.sample([i for i in range(m)], num_cols_remove)
        A = remove_cols(A, cols_to_del)

        m, n = A.shape
        p = Matrix(m, 1, lambda i, j: ri(2, 99))
        x = Matrix(n, 1, lambda i, j: ri(-5, 5))

        b = vector_mod_vector(A * x, p)

        x_sol = solve(A, b, p)
        assert (vector_mod_vector(A * x_sol, p) == b)

    def test_lower_rank_mapping(self):
        """
        Test the equation solver on a matrix A
        which is neither surjective nor injective.
        """
        # Create a matrix of lower rank
        n = self.n
        r = self.r
        g = Matrix(r, n, lambda i, j : ri(-9, 9))
        A = g.T * g

        # Set up equations
        x = Matrix(n, 1, lambda i, j: ri(-5, 5))
        p = Matrix(n, 1, lambda i, j: ri(10, 100))
        b = vector_mod_vector(A * x, p)

        # Solve the equation and verify
        x_sol = solve(A, b, p)
        assert (vector_mod_vector(A * x_sol, p) == b)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains stochastic tests kernel, cokernel,
image and coimage algorithms based on the SVD for maps
between vector spaces over R.
"""

from sympy import Matrix
from random import choice, random
from abelian.linalg.factorizations_reals import real_kernel, real_cokernel, \
    real_coimage, real_image

class TestFactorizationsReals:

    @classmethod
    def setup_class(cls):
        """
        Set up a random matrix.
        """
        # Add more zero entries to cover edge cases with higher probability
        numbers = [0] * 10 + list(range(-5, 5)) + [random() for i in range(8)]
        cls.m, cls.n = choice([1,2,3,4]), choice([1,2,3,4])
        cls.A = Matrix(cls.m, cls.n, lambda i, j: choice(numbers))

    def test_kernel(self):
        """
        Test the factorization.
        """

        # Take the kernel
        A = self.A
        m, n = A.shape
        ker = real_kernel(A)

        # Verify norm and dimension
        A_ker_product = A * ker
        small_norm = sum(abs(k) for k in A_ker_product) < 10e-10
        correct_dim = A_ker_product.rows == m

        assert small_norm and correct_dim

    def test_cokernel(self):
        """
        Test the kernel factorization.
        """

        # Take the kernel
        A = self.A
        m, n = A.shape
        coker = real_cokernel(A)

        # Verify norm and dimension
        coker_A_product = coker * A
        small_norm = sum(abs(k) for k in coker_A_product) < 10e-10
        correct_dim = coker_A_product.cols == n

        assert small_norm and correct_dim

    def test_image_coimage(self):
        """
        Test the image/coimage factorization.
        """
        # Take the kernel
        A = self.A
        im = real_image(A)
        coim = real_coimage(A)

        # Verify norm and dimension
        im_coim_product = im * coim
        small_norm = sum(abs(k) for k in (A - im_coim_product)) < 10e-10

        assert small_norm


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains functions which calculate mapping properties of
homomorphisms between R^n and R^m using the singular value decomposition (SVD).
All the inputs and outputs are of type
:py:class:`~sympy.matrices.dense.MutableDenseMatrix`.
"""

from sympy import Matrix, diag
from abelian.linalg.utils import columns_as_list
import numpy as np

def numerical_SVD(A):
    """
    Compute U,S,V such that U*S*V = A.

    The input is converted to numerical data, the SVD is computed using
    the np.linalg.svd routine, which wraps the LAPACK routine _gesdd.
    The data is then converted to a sympy matrix and returned.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix.

    Returns
    -------
    U : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A (close to) orthogonal sympy matrix.

    S : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A diagonal sympy matrix matrix.

    V : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A (close to) orthogonal sympy matrix.

    Examples
    ---------
    >>> A = Matrix([[1, 2], [3, 4]])
    >>> U, S, V = numerical_SVD(A)
    >>> # U is orthogonal (up to machine precision or so)
    >>> abs(abs(U.det()) - 1) < 10e-10
    True
    >>> # Verify that the decomposition is close to the original
    >>> sum(abs(k) for k in (U*S*V - A))  < 10e-10
    True
    """
    # Convert A to a numerical type instead of symbolic
    A_numeric = np.array(columns_as_list(A), dtype=float).T

    # Take the singular value decomposition
    U, S, V = np.linalg.svd(A_numeric, full_matrices=True)

    # Convert to sympy data types
    U, S, V = Matrix(U.tolist()), diag(*S.tolist()), Matrix(V.tolist())
    return U, S, V

def numerical_rank(A):
    """
    Convert to numerical matrix and compute rank.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix.

    Returns
    -------
    r : int
        The rank of A.

    Examples
    ---------
    >>> A = Matrix([[1, 2], [3, 4]])
    >>> numerical_rank(A)
    2
    >>> A = Matrix([[0, 0], [0, 10e-10]])
    >>> numerical_rank(A)
    1
    """
    # Convert A to a numerical type instead of symbolic
    A_numeric = np.array(columns_as_list(A), dtype=float).T

    # Compute the rank and return
    return np.linalg.matrix_rank(A_numeric)

def real_kernel(A):
    """
    Find the kernel of A, when the entries are real.

    Converts the matrix to a numerical input, computes the SVD,
    finds the kernel monomorphism (null space of A), converts back
    to a sympy-matrix and returns.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix.

    Returns
    -------
    K : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The kernel of A.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[1, 0, 1],
    ...             [0, 1, 1],
    ...             [2, 2, 4]])
    >>> ker = real_kernel(A)
    >>> # Verify the decomposition
    >>> sum(abs(k) for k in (A * ker)) < 10e-15
    True
    """

    # Get size, rank and decompose A
    m, n = A.shape
    r = numerical_rank(A)
    U, S, V = numerical_SVD(A)

    # Take transpose (inverse) of V
    V = V.T

    # Return the last (n-r) columns of V
    if (n-r) != 0:
        return V[:, -(n-r):]
    else:
        # Return an empty matrix along one dimension
        return V[:, :0]

def real_cokernel(A):
    """
    Find the cokernel of A, when the entries are real.

    Converts the matrix to a numerical input, computes the SVD,
    finds the cokernel epimorphism (null space of A^T), converts back
    to a sympy-matrix and returns.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix.

    Returns
    -------
    K : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The cokernel of A.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[1, 0],
    ...             [0, 1],
    ...             [2, 2]])
    >>> coker = real_cokernel(A)
    >>> # Verify the decomposition
    >>> sum(abs(k) for k in (coker * A)) < 10e-15
    True
    """

    # Get size, rank and decompose A
    m, n = A.shape
    r = numerical_rank(A)
    U, S, V = numerical_SVD(A)

    # Take the transpose/inverse
    U = U.T

    # Return the last (m-r) rows of A
    if (m - r) != 0:
        return U[-(m - r):, :]
    else:
        return U[:0, :]

def real_image(A):
    """
    Find the image of A, when the entries are real.

    Converts the matrix to a numerical input, computes the SVD,
    finds the image monomorphism (column space), converts back
    to a sympy-matrix and returns.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix.

    Returns
    -------
    K : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The image of A.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[1, 0],
    ...             [0, 1],
    ...             [1, 1]])
    >>> im = real_image(A)
    >>> coim = real_coimage(A)
    >>> # Verify the decomposition
    >>> sum(abs(k) for k in (A - im * coim)) < 10e-15
    True
    """

    # Get the rank and decompose A, return first r columns of U
    r = numerical_rank(A)
    U, S, V = numerical_SVD(A)
    return U[:, :r] * S[:r, :r]


def real_coimage(A):
    """
    Find the coimage of A, when the entries are real.

    Converts the matrix to a numerical input, computes the SVD,
    finds the coimage epimorphism (row space of A), converts back
    to a sympy-matrix and returns.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy matrix.

    Returns
    -------
    K : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The coimage of A.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[1, 0, 0],
    ...             [0, 1, 0]])
    >>> im = real_image(A)
    >>> coim = real_coimage(A)
    >>> # Verify the decomposition
    >>> sum(abs(k) for k in (A - im * coim)) < 10e-15
    True
    """

    # Get rank, decompose A and return the first r rows of V
    r = numerical_rank(A)
    U, S, V = numerical_SVD(A)
    return V[:r, :]


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose = False)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
from sympy import Matrix, diag
from random import randint as ri
from abelian import HomLCA
from abelian.utils import mod
from abelian.linalg.utils import order_of_vector, remove_zero_columns, \
    nonzero_diag_as_list, vector_mod_vector
from abelian.linalg.factorizations import smith_normal_form
import itertools
from functools import reduce
from operator import mul
import operator
import functools

def random_zero_heavy(low, high):
    """
    Draw a random number, with approx 50% probability of zero.
    """
    return random.choice(list(range(low, high)) + [0]*(high - low))

class TestFGAFactorizationsNaively:

    @classmethod
    def setup_class(cls):
        max_size = 15
        m, n = ri(1, max_size), ri(2, max_size)
        print(m, n)
        num_generators = ri(1, 2)
        generators = [[mod(ri(1, max_size), m), mod(ri(1, max_size), n)]
                      for i in range(num_generators)]
        print(generators)
        periods = [order_of_vector(g, [m, n]) for g in generators]
        print(periods)

        cls.generators = generators
        cls.periods = periods
        cls.m, cls.n = m, n

    def test_cokernel_target(self):
        """
        Test the quotient group.
        """

        # According to proof, this is true
        A = Matrix(self.generators).T
        ker_pi = remove_zero_columns(diag(self.m, self.n))
        U, S, V = smith_normal_form(A.row_join(ker_pi))
        diagonal = nonzero_diag_as_list(S)
        (m, n), r = A.shape, len(diagonal)
        quotient = diagonal + [0] * (m - r)
        print(quotient)

        # Compute the quotient naively. Iterate through all combinations of
        # the generators, store the group elements which we find, and compare
        # this with the order of the group.
        seen_elements = set()
        periods_vector = Matrix([self.m, self.n])
        for p in itertools.product(*[range(p) for p in self.periods]):
            element = vector_mod_vector(A*Matrix(p), periods_vector)
            seen_elements.add(tuple(element))

        naive_quotient = self.m * self.n / len(seen_elements)
        assert naive_quotient % 1 == 0

        assert naive_quotient == functools.reduce(operator.mul, quotient)




if __name__ == '__main__':
    test = TestFGAFactorizationsNaively()
    test.setup_class()
    test.test_cokernel_target()


#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""
This module consists of classes representing homomorphisms between
elementary LCAs, the HomLCA class.
"""

from collections.abc import Callable
from sympy import Matrix, diag, latex, Integer, gcd
from abelian.groups import LCA
from abelian.linalg.utils import remove_zero_columns, nonzero_diag_as_list, \
    matrix_mod_vector, order_of_vector, remove_cols, remove_rows, \
    columns_as_list, reciprocal_entrywise, diag_times_mat, mat_times_diag
from abelian.linalg.factorizations import smith_normal_form
from abelian.linalg.factorizations_reals import real_image, real_coimage, \
    real_cokernel, real_kernel
from abelian.linalg.solvers import solve_epi
from abelian.functions import LCAFunc


class HomLCA(Callable):
    """
    A homomorphism between elementary LCAs.
    """

    # The types allowed as entries in A
    _A_all_entry_types = (int, float, complex)
    _A_integer_entry_types = (int, Integer)

    def __init__(self, A, target = None, source = None):
        """
        Initialize a homomorphism.

        Parameters
        ----------
        A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix` or list
            A sympy matrix representing the homomorphism. The user may also
            use a list of lists in the form [row1, row2, ...] as input.
        target : LCA or list
            The target of the homomorphism. If None, a discrete target of
            infinite order is used as the default.
        source : LCA or list
            The source of the homomorphism. If None, a discrete source of
            infinite order is used as the default.

        Examples
        ---------
        >>> # If no source/target is given, a free discrete group is assumed
        >>> phi = HomLCA([[1,2],
        ...               [3,4]])
        >>> phi.source.is_FGA() and phi.target.is_FGA()
        True

        >>> # If no source is given, a free discrete group is assumed
        >>> phi = HomLCA([[1,2],
        ...               [3,4]], target = [5, 5])
        >>> phi.source.is_FGA()
        True

        >>> # The homomorphism must be valid
        >>> from abelian import LCA, HomLCA
        >>> T = LCA(orders = [1], discrete = [False])
        >>> R = LCA(orders = [0], discrete = [False])
        >>> phi = HomLCA([1], target = R, source = T)
        Traceback (most recent call last):
            ...
        ValueError: 1: [T] -> [R] is not homomorphism


        """
        A, target, source = self._verify_init(A, target, source)
        assert isinstance(A, Matrix)
        assert isinstance(target, LCA)
        assert isinstance(source, LCA)
        self.A = A
        self.target = target
        self.source = source

        # TODO : Should we project to target automatically?

    @classmethod
    def identity(cls, group):
        """
        Return the identity morphism.

        Examples
        ---------
        >>> from abelian import LCA, HomLCA
        >>> H = LCA([5, 6, 7])
        >>> G = LCA([0, 0])
        >>> phi = HomLCA([[1,2], [3,4], [5,6]], source = G, target = H)
        >>> Id_H = HomLCA.identity(H)
        >>> Id_G = HomLCA.identity(G)
        >>> # Verify the properties of the identity morphism
        >>> Id_H * phi == phi
        True
        >>> phi * Id_G == phi
        True
        """
        m = group.length()
        return cls(Matrix.eye(m), source = group, target = group)

    @staticmethod
    def _is_scalar_homomorphism(scalar, target, source):
        """
        Check if the source, target and scalar value is a homomorphism.

        Parameters
        ----------
        scalar : float
            A scalar value.
        target : LCA
            An LCA of length 1: either R, T, Z or Z_n.
        source : LCA
            An LCA of length 1: either R, T, Z or Z_n.

        Examples
        ---------
        >>> from abelian import LCA
        >>> Z = LCA([0])
        >>> # This defines a homomorphism
        >>> HomLCA._is_scalar_homomorphism(2, Z, Z)
        True
        >>> # This does not
        >>> HomLCA._is_scalar_homomorphism(2.5, Z, Z)
        False
        >>> # Two examples with FGAs with finite order
        >>> Z_n = LCA([3])
        >>> Z_m = LCA([4])
        >>> HomLCA._is_scalar_homomorphism(4, target = Z_m, source = Z_n)
        True
        >>> HomLCA._is_scalar_homomorphism(3, target = Z_m, source = Z_n)
        False
        """

        # Verify that both groups are of length 1
        if len(source) != 1 or len(target) != 1:
            raise ValueError('Groups must have length 1.')

        # Set up groups here for readability later
        R = LCA([0], [False])
        T = LCA([1], [False])
        Z = LCA([0], [True])
        Z_1 = LCA([1], [True])

        # Trivial homomorphism is always valid
        if scalar == 0 or source == Z_1:
            return True

        # Defined for readability
        def integer_value(arg):
            return arg % 1 == 0

        # The source group is R
        if source == R:
            if target == R or target == T:
                return True
            else:
                return False

        # The source group is T
        if source == T:
            if target == T and integer_value(scalar):
                return True
            else:
                return False

        # The source group is Z
        if source == Z:
            if target == R or target == T:
                return True
            if integer_value(scalar):
                return True
            return False

        # The source group is Z_n
        n = source.orders[0]
        if source.discrete[0] is True and n > 0:
            if target == R or target == Z:
                return False

            if target == T and integer_value(scalar * n):
                return True

            m = target.orders[0]
            if target.discrete[0] is True and m > 0:
                if integer_value((scalar * gcd(m, n) / m)):
                    return True
                else:
                    return False

        return False


    @classmethod
    def _verify_init(cls, A, target, source):
        """
        Verify the inputs.

        Parameters
        ----------
        A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix` or list
            A sympy matrix representing the homomorphism. The user may also
            use a list of lists in the form [row1, row2, ...] as input.
        target : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`,
        list or LCA
            The target of the homomorphism. If None, a discrete target of
            infinite order is used as the default.
        source : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`,
        list or LCA
            The source of the homomorphism. If None, a discrete source of
            infinite order is used as the default.

        Returns
        -------
        A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        target : LCA
        source : LCA
        """

        # If the input matrix is a list of lists, convert to matrix
        if isinstance(A, list):
            A = Matrix(A)
        elif isinstance(A, cls._A_all_entry_types):
            A = Matrix([A])


        # If no target is given, assume free-to-free morphism
        if (target is None):
            target = [0] * A.rows
            source = [0] * A.cols

        # If target is given, and no source, assume left-free morphism
        elif (source is None):
            source = [0] * A.cols

        # If lists of matrices were passed, convert to LCAs
        if isinstance(target, (list, Matrix)):
            target = LCA(list(target))

        if isinstance(source, (list, Matrix)):
            source = LCA(list(source))

        # If the matrix is empty on either side, replace the LCA
        m, n = A.shape
        if m == 0:
            return A, LCA([]), source
        if n == 0:
            return A, target, LCA([])



        # Verify that the input is actually a homomorphism
        for j, s in enumerate(source):
            for i, t in enumerate(target):
                if not cls._is_scalar_homomorphism(A[i, j], t, s):
                    data = A[i, j], s, t
                    err = "{}: {} -> {} is not homomorphism".format(*data)
                    raise ValueError(err)


        return A, target, source


    def __add__(self, other):
        """
        Override the addition (`+`) operator,
        see :py:meth:`~abelian.morphisms.HomLCA.add`.
        """
        return self.add(other)


    def __call__(self, source_element):
        """
        Override function calls,
        see :py:meth:`~abelian.morphisms.HomLCA.evaluate`.
        """
        return self.evaluate(source_element)


    def __eq__(self, other):
        """
        Override the equality (`==`) operator,
        see :py:meth:`~abelian.morphisms.HomLCA.equal`.
        """
        return self.equal(other)

    def __getitem__(self, args):
        """
        Override the slice operator,
        see :py:meth:`~abelian.morphisms.HomLCA.getitem`.
        """
        return self.getitem(args)


    def __mul__(self, other):
        """
        Override the `*` operator,
        see :py:meth:`~abelian.morphisms.HomLCA.compose`.
        """
        return self.compose(other)

    def __pow__(self, power, modulo=None):
        """
        Override the pow (`**`) operator,
        see :py:meth:`~abelian.morphisms.HomLCA.compose_self`.
        """
        return self.compose_self(power)

    def __radd__(self, other):
        """
        Override the addition (`+`) operator,
        see :py:meth:`~abelian.morphisms.HomLCA.add`.
        """
        return self.add(other)


    def __repr__(self):
        """
        Override the ``repr()`` function.
        """
        rep = 'source: {}     target: {}\n'.format(repr(self.source),
                                                   repr(self.target))
        rep += str(self.A)
        return rep


    def __rmul__(self, other):
        """
        Override the `*` operator,
        see :py:meth:`~abelian.morphisms.HomLCA.compose`.
        """
        return self.compose(other)

    def add(self, other):
        """
        Elementwise addition.

        Elementwise addition of the underlying matrix.

        Parameters
        ----------
        other : HomLCA or numeric
            A homomorphism to add to the current one, or a number.

        Returns
        -------
        homomorphism : HomLCA
            A new homomorphism with the argument added.

        """
        # If the `other` argument is a numeric type
        if isinstance(other, self._A_all_entry_types):
            m, n = self.A.shape
            new_A = self.A + Matrix(m, n, lambda i,j : other)
            return type(self)(new_A, target=self.target, source=self.source)

        # If the `other` argument is a numeric type
        if isinstance(other, type(self)):
            if not self.target == other.target and self.source == other.source:
                raise ValueError('Sources and targets must match..')

            new_A = self.A + other.A
            return type(self)(new_A, target=self.target, source=self.source)

        format_args = type(self), type(other)
        raise ValueError('Cannot add {} and {}'.format(*format_args))


    def compose(self, other):
        """
        Compose two homomorphisms.

        The composition of `self` and `other` is first other, then self.

        Parameters
        ----------
        other : HomLCA
            The homomorphism to compose with.

        Returns
        -------
        homomorphism : HomLCA
            The composition of `self` and `other`, i.e. `self` ( `other` (x)).

        Examples
        --------
        >>> phi = HomLCA([[1, 0, 1],
        ...               [0, 1, 1]])
        >>> ker_phi = HomLCA([1, 1, -1])
        >>> (phi * ker_phi) == HomLCA([0, 0])
        True
        >>> phi.compose(ker_phi) == HomLCA([0, 0])
        True
        """

        # If the `other` argument is a numeric type
        if isinstance(other, self._A_all_entry_types):
            new_A = self.A * other
            return type(self)(new_A, target=self.target, source=self.source)

        # If the `other` argument is a HomLCA
        if isinstance(other, type(self)):
            if not other.target == self.source:
                raise ValueError('Target of other must equal source of self.')

            new_A = self.A * other.A
            return type(self)(new_A, target=self.target, source=other.source)

        # If the composition is an HomLCA, then a LCAFunc,
        # it's valid if the domain matches
        if isinstance(other, LCAFunc):
            if other.domain.equal(self.target):
                return other.pullback(self)

        format_args = type(self), type(other)
        raise ValueError('Cannot compose/add {} and {}'.format(*format_args))

    def compose_self(self, power):
        """
        Repeated composition of an endomorphism.

        Parameters
        ----------
        power : int
            The number of times to compose with self.

        Returns
        -------
        homomorphism : HomLCA
            The endomorphism composed with itself `power` times.

        Examples
        --------
        >>> from sympy import diag
        >>> phi = HomLCA(diag(2, 3))
        >>> phi**3 == HomLCA(diag(2**3, 3**3))
        True
        """
        product = self.copy()
        for prod in range(power - 1):
            product = product.compose(self)
        return product

    def copy(self):
        """
        Return a copy of the homomorphism.

        Returns
        -------
        homomorphism : HomLCA
            A copy of the homomorphism.

        Examples
        --------
        >>> phi = HomLCA([1, 2, 3])
        >>> phi.copy() == phi
        True
        """
        return type(self)(self.A, target = self.target, source = self.source)

    def dual(self):
        """
        Compute the dual homomorphism.

        TODO: Write detailed description.


        Returns
        -------
        dual : HomLCA
            The dual homomorphism.

        Examples
        ---------
        >>> phi = HomLCA([2])
        >>> phi_dual = phi.dual()
        >>> phi_dual.source == phi_dual.target
        True

        Computing duals by first calculating orders

        >>> # Project, then find dual
        >>> phi = HomLCA([2], target = [10])
        >>> phi_proj = phi.project_to_source()
        >>> phi_project_dual = phi_proj.dual()
        >>> phi_project_dual == HomLCA([1], [5], [10])
        True
        >>> # Do not project
        >>> phi_dual = phi.dual()
        >>> phi_dual == HomLCA([1/5], LCA([1], [False]), [10])
        True
        """

        # Flip the source and target for the dual morphism
        dual_source, dual_target = self.target.dual(), self.source.dual()

        # Calculate the matrix representing the dual homomorphism
        diag_p = Matrix([1 if e == 0 else e for e in self.source.orders])
        diag_q_inv = reciprocal_entrywise(Matrix([1 if e == 0 else e for e
                                                  in self.target.orders]))
        dual_A = mat_times_diag(diag_times_mat(diag_p, self.A.T), diag_q_inv)

        # Create new FGA object and return
        return HomLCA(dual_A, target = dual_target, source = dual_source)

    def equal(self, other):
        """
        Whether or not two homomorphisms are equal.

        Two HomLCAs are equal iff (1) the sources are equal, (2) the targets
        are equal and (3) the matrices representing the homomorphisms are
        equal.

        Parameters
        ----------
        other : HomLCA
            A HomLCA to compare equality with.

        Returns
        -------
        equal : bool
            Whether or not the HomLCAs are equal.

        Examples
        ---------
        >>> phi = HomLCA([1], target=[0], source = [0]) # Explicit
        >>> psi = HomLCA([1])   # Shorter, defaults to the above
        >>> phi == psi
        True

        """
        source_equal = self.source == other.source
        target_equal = self.target == other.target
        A_equal = self.A == other.A
        return source_equal and target_equal and A_equal

    def evaluate(self, source_element):
        """
        Apply the homomorphism to an element.

        Parameters
        ----------
        source_element

        Returns
        -------

        Examples
        ---------
        >>> from sympy import diag
        >>> phi = HomLCA(diag(3, 4), target = [5, 6])
        >>> phi.evaluate([2, 3])
        [1, 0]
        >>> phi.evaluate(Matrix([2, 3]))
        Matrix([
        [1],
        [0]])
        """

        # Project the element to the source LCA of the homomorphism
        source_element = self.source.project_element(source_element)

        m, n = self.A.shape
        if n == 0:
            evaluated = [0]

        # Apply the homomorphism using matrix multiplication, depending
        # on the type of the input
        elif isinstance(source_element, Matrix):
            evaluated = self.A * source_element
        else:
            evaluated = self.A * Matrix(source_element)

        # Project to the target LCA of the homomorphism
        projected = self.target.project_element(evaluated)

        # Return the same type of data as the input
        if isinstance(source_element, Matrix):
            return projected
        else:
            rows = columns_as_list(projected.T)
            return [r[0] for r in rows]

    def getitem(self, args):
        """
        Return a slice of the homomorphism.

        Slices the object with the common matrix slice notation, e.g. A[rows,
        cols], where the `rows` and `cols` objects can be either
        integers or slice objects. If the homomorphism is represented by a
        column or row matrix, then the notation A[key] will also work.
        The underlying matrix and the source and target LCAs are all sliced.

        Parameters
        ----------
        args : slice
            A slice or a tuple with (slice_row, slice_col).

        Returns
        --------
        homomorphism : HomLCA
            A sliced homomorphism.

        Examples
        --------

        The homomorphism is sliced using two input arguments.

        >>> from sympy import diag
        >>> phi = HomLCA(diag(4,5,6))
        >>> phi[0,:] == HomLCA([[4, 0, 0]])
        True
        >>> phi[:,1] == HomLCA([0, 5, 0])
        True

        If the homomorphism is represented by a row or column, one arg will do.

        >>> phi = HomLCA([1,2,3])
        >>> phi[0:2] == HomLCA([1,2])
        True
        """
        # Two arguments passed, slice row and column of A
        if isinstance(args, (slice, int)):
            # One argument was passed
            rows_A, cols_A = self.A.shape
            slice1 = args
            if cols_A == 1:
                # A column vector
                new_A = self.A[slice1]
                new_target = self.target[slice1]
                new_source = self.source
                return type(self)(new_A, target=new_target, source=new_source)
            elif rows_A == 1:
                # A column vector
                new_A = self.A[slice1]
                new_target = self.target
                new_source = self.source[slice1]
                return type(self)(new_A, target=new_target, source=new_source)
            else:
                error = 'One slice argument was passed, but the morphism has' \
                        ' dimension {} x {}'.format(rows_A, cols_A)
                return ValueError(error)


        elif isinstance(args, tuple) and len(args) == 2:

            # Two arguments were passed, slice the matrix and rows/cols
            slice1, slice2 = args
            if not (isinstance(slice1, (slice, int)) and
                isinstance(slice1, (slice, int))):
                raise ValueError('Arguments must be slice() or integer.')

            new_A = self.A[slice1, slice2]
            new_target = self.target[slice1]
            new_source = self.source[slice2]
            return type(self)(new_A, target=new_target, source=new_source)

        else:
            raise ValueError('slice() takes 1 or 2 arguments.')


    def remove_trivial_groups(self):
        """
        Remove trivial groups.

        A group is trivial if it is discrete with order 1, i.e. Z_1.
        Removing trivial groups from the target group means removing the
        Z_1 groups from the target, along with the corresponding rows of
        the matrix representing the homomorphism.
        Removing trivial groups from the source group means removing the
        groups Z_1 from the source, i.e. removing every column (generator)
        with order 1.

        Returns
        -------
        homomorphism : HomLCA
            A homomorphism where the trivial groups have been removed from
            the source and the target. The corresponding rows and columns of
            the matrix representing the homomorphism are also removed.

        Examples
        --------
        >>> target = [1, 7]
        >>> phi = HomLCA([[2, 1], [7, 2]], target=target)
        >>> projected = HomLCA([[2]], target=[7], source = [7])
        >>> phi.project_to_source().remove_trivial_groups() == projected
        True

        """
        def trivial(order, discrete):
            return discrete and (order == 1)

        # Get indices where the value of the source is 1
        generator = enumerate(self.source._iterate_tuples())
        cols_to_del = [i for (i, (d, p)) in generator if trivial(d, p)]
        new_A = remove_cols(self.A, cols_to_del)

        # Get indices where the value of the target is 1
        generator = enumerate(self.target._iterate_tuples())
        rows_to_del = [i for (i, (d, p)) in generator if trivial(d, p)]
        new_A = remove_rows(new_A, rows_to_del)

        new_source = self.source.remove_indices(cols_to_del)
        new_target = self.target.remove_indices(rows_to_del)

        return type(self)(new_A, new_target, new_source)


    def stack_diag(self, other):
        """
        Stack diagonally.

        Parameters
        ----------
        other : HomLCA
            A homomorphism to stack with the current one.

        Returns
        -------
        stacked_vert : HomLCA
            The result of stacking the homomorphisms on diagonally.

        Examples
        --------
        >>> phi = HomLCA([1])
        >>> psi = HomLCA([2])
        >>> phi.stack_diag(psi) == HomLCA([[1, 0], [0, 2]])
        True

        """
        new_source = self.source + other.source
        new_target = self.target + other.target
        new_A = diag(self.A, other.A)

        return type(self)(new_A, target = new_target, source = new_source)

    def stack_horiz(self, other):
        """
        Stack horizontally (column wise).

        The targets must be the same, the sources will be concatenated.
        The stacking is done to create a matrix with structure [self,
        other], i.e. "Putting `self` to the left of `other`."

        Parameters
        ----------
        other : HomLCA
            A homomorphism to stack with the current one.

        Returns
        -------
        stacked_vert : HomLCA
            The result of stacking the homomorphisms side by side.

        Examples
        --------
        >>> phi = HomLCA([1])
        >>> psi = HomLCA([2])
        >>> phi.stack_horiz(psi) == HomLCA([[1, 2]])
        True

        """
        if not self.target == other.target:
            raise ValueError('Targets must be equal to stack horizontally.')
        new_source = self.source + other.source
        new_target = self.target
        new_A = self.A.row_join(other.A)
        return type(self)(new_A, target = new_target, source = new_source)


    def stack_vert(self, other):
        """
        Stack vertically (row wise).

        The sources must be the same, the targets will be concatenated.
        The stacking is done to create a matrix with structure [[self],
        [other]], i.e. "Putting `self` on top of of `other`."

        Parameters
        ----------
        other : HomLCA
            A homomorphism to stack with the current one.

        Returns
        -------
        stacked_vert : HomLCA
            The result of stacking the homomorphisms on top of each other.

        Examples
        --------
        >>> phi = HomLCA([1])
        >>> psi = HomLCA([2])
        >>> phi.stack_vert(psi) == HomLCA([1, 2])
        True

        """
        if not self.source == other.source:
            raise ValueError('Sources must be equal to stack vertically.')
        new_source = self.source
        new_target = self.target + other.target
        new_A = self.A.col_join(other.A)
        return type(self)(new_A, target = new_target, source = new_source)

    def update(self, new_A = None, new_target = None, new_source = None):
        """
        Return a new homomorphism with updated properties.
        """

        if new_A is None:
            new_A = self.A
        if new_target is None:
            new_target = self.target
        if new_source is None:
            new_source = self.source

        return type(self)(new_A, target=new_target, source=new_source)


    def _is_homFGA(self):
        """
        Whether or not is a homomorphism between FGAs.

        Returns
        -------
        homFGA : bool
            Whether or not it's a homFGA.

        Examples
        ---------
        >>> phi = HomLCA([1], source = [1], target = [1])
        >>> phi._is_homFGA()
        True

        """
        integer_entries = all([i % 1 == 0 for i in self.A])
        if self.source.is_FGA() and self.target.is_FGA() and integer_entries:
            return True
        else:
            return False

    def to_latex(self):
        """
        Return the homomorphism as a :math:`\LaTeX` string.

        Returns
        -------
        latex : str
            The HomLCA formatted as a LaTeX string.

        Examples
        --------
        >>> phi = HomLCA([1])
        >>> phi.to_latex()
        '\\\\begin{pmatrix}1\\\\end{pmatrix}:\\\\mathbb{Z} \\\\to \\\\mathbb{Z}'
        """
        latex_code = latex(self.A)
        latex_code = latex_code.replace(r'\left[\begin{matrix}',
                                        r'\begin{pmatrix}')
        latex_code = latex_code.replace(r'\end{matrix}\right]',
                                        r'\end{pmatrix}')
        format_args = self.source.to_latex(), self.target.to_latex()
        latex_code += r':{} \to {}'.format(*format_args)
        return latex_code


    @classmethod
    def zero(cls, target, source):
        """
        Initialize the zero morphism.

        Parameters
        ----------
        target : LCA or list
            The target of the homomorphism. If None, a discrete target of
            infinite order is used as the default.
        source : LCA or list
            The source of the homomorphism. If None, a discrete source of
            infinite order is used as the default.

        Examples
        ---------
        >>> zero = HomLCA.zero([0]*3, [0]*3)
        >>> zero([1, 5, 7]) == [0, 0, 0]
        True
        """
        rows = len(target)
        cols = len(source)
        A = Matrix(rows, cols, lambda i, j : 0)
        return cls(A, target = target, source = source)

    @property
    def shape(self):
        """
        The shape (`rows`, `cols`).

        Returns
        -------
        shape : tuple
            A tuple with the shape of the underlying matrix A, i.e. (`rows`,
            `cols`).

        """
        return self.A.shape


    def annihilator(self):
        """
        Compute the annihilator monomorphism.
        """
        # TODO: Write this method.
        return self.cokernel().dual().remove_trivial_groups()

    def coimage(self):
        """
        Compute the coimage epimorphism.

        Returns
        -------
        homomorphism : HomLCA
            The coimage homomorphism.

        Examples
        --------
        >>> phi = HomLCA([[4, 4],
        ...               [2, 8]], target = [16, 16])
        >>> im = phi.image().remove_trivial_groups()
        >>> coim = phi.coimage().remove_trivial_groups()
        >>> phi == (im * coim).project_to_target()
        True

        """
        R = LCA([0], [False])
        target_real_noncompact = all(g == R for g in self.target)
        A_is_integer = all(isinstance(a, self._A_integer_entry_types) for a
                           in self.A)


        # If the mapping is R^d -> R^d, the coimage is found with the SVD
        if all(g == R for g in self.source) and \
            all(g == R for g in self.target):
            coimage_A = real_coimage(self.A)
            m, n = coimage_A.shape
            return type(self)(coimage_A, source=R**n, target=R**m)

        if self.source.is_FGA() and target_real_noncompact and A_is_integer:
            return self._FGA_coimage()

        if self._is_homFGA():
            return self._FGA_coimage()
        else:
            raise NotImplementedError('Not implemented.')

    def cokernel(self):
        """
        Compute the cokernel epimorphism.

        Returns
        -------
        homomorphism : HomLCA
            The cokernel homomorphism.

        Examples
        --------
        >>> phi = HomLCA([[1, 0], [0, 1], [1, 1]])
        >>> coker = phi.cokernel()
        >>> coker.target.isomorphic(LCA([1, 1, 0]))
        True

        """


        if self._is_homFGA():
            return self._FGA_cokernel()

        # If it's a map Z^m -> R^n, then return inverse matrix
        Z = LCA(orders = [0], discrete = [True])
        R = LCA(orders=[0], discrete=[False])
        T = LCA(orders=[1], discrete=[False])
        m, n = self.shape

        # If the mapping is R^d -> R^d, the kernel is found with the SVD
        if all(g == R for g in self.source) and \
            all(g == R for g in self.target):
            cokernel_A = real_cokernel(self.A)
            m, n = cokernel_A.shape
            return type(self)(cokernel_A, source=R**n, target=R**m)


        if (m == n) and self.source == Z**m and self.target == R**m:
            inverse_A = self.A.inv()
            return type(self)(inverse_A, target = T**m, source = self.target)

        else:
            raise NotImplementedError('Not implemented.')

    def image(self):
        """
        Compute the image monomorphism.

        Returns
        -------
        homomorphism : HomLCA
            The image homomorphism.

        Examples
        --------
        >>> phi = HomLCA([[4, 4],
        ...               [2, 8]], target = [64, 32])
        >>> im = phi.image().remove_trivial_groups()
        >>> coim = phi.coimage().remove_trivial_groups()
        >>> phi == (im * coim).project_to_target()
        True

        >>> # Image computations are also allowed when target is R
        >>> R = LCA(orders = [0], discrete = [False])
        >>> sample_matrix = [[1, 2, 3], [2, 3, 5]]
        >>> phi_sample = HomLCA(sample_matrix, target = R + R)
        >>> phi_sample_im = phi_sample.image().remove_trivial_groups()
        >>> phi_sample_im == phi_sample[:, 1:]
        True

        """
        R = LCA([0], [False])
        target_real_noncompact = all(g == R for g in self.target)
        A_is_integer = all(isinstance(a, self._A_integer_entry_types) for a
                           in self.A)


        # If the mapping is R^d -> R^d, the image is found with the SVD
        if all(g == R for g in self.source) and \
            all(g == R for g in self.target):
            image_A = real_image(self.A)
            m, n = image_A.shape
            return type(self)(image_A, source=R**n, target=R**m)


        if self.source.is_FGA() and target_real_noncompact and A_is_integer:
            return self._FGA_image()
        if self._is_homFGA():
            return self._FGA_image()
        else:
            raise NotImplementedError('Not implemented.')


    def kernel(self):
        """
        Compute the kernel monomorphism.

        Returns
        -------
        homomorphism : HomLCA
            The kernel homomorphism.

        Examples
        --------
        >>> phi = HomLCA([[1, 0, 1], [0, 1, 1]])
        >>> phi.kernel() == HomLCA([-1, -1, 1])
        True

        """
        R = LCA([0], [False])
        T = LCA([1], [False])
        target_real_noncompact = all(g == R for g in self.target)
        A_is_integer = all(isinstance(a, self._A_integer_entry_types) for a
                           in self.A)

        # If the mapping is R^d -> T^d, the kernel is the inverse
        m, n = self.shape
        if m == n:
            if self.source == R**m and self.target == T**m:
                kernel_source = LCA([0], [True])**m
                A_inv = self.A.inv()
                return type(self)(A_inv, source = kernel_source, target =
                self.source)

        # If the mapping is R^d -> R^d, the kernel is found with the SVD
        if all(g == R for g in self.source) and \
            all(g == R for g in self.target):
            kernel_A = real_kernel(self.A)
            m, n = kernel_A.shape
            return type(self)(kernel_A, source=R**n, target=R**m)


        if self.source.is_FGA() and target_real_noncompact and A_is_integer:
            return self._FGA_kernel()

        if self._is_homFGA():
            return self._FGA_kernel()
        else:
            raise NotImplementedError('Not implemented.')

    def _FGA_coimage(self):
        """
        Compute the coimage epimorphism.

        Returns
        -------
        homomorphism : HomLCA
            The coimage homomorphism.

        Examples
        --------
        >>> phi = HomLCA([[4, 4],
        ...               [2, 8]], target = [16, 16])
        >>> im = phi.image().remove_trivial_groups()
        >>> coim = phi.coimage().remove_trivial_groups()
        >>> phi == (im * coim).project_to_target()
        True

        """
        # Compute the cokernel of the kernel and return
        kernel = self.kernel()
        coimage = kernel.cokernel()
        return coimage

    def _FGA_cokernel(self):
        """
        Compute the cokernel epimorphism.

        Returns
        -------
        homomorphism : HomLCA
            The cokernel homomorphism.

        Examples
        --------
        >>> phi = HomLCA([[1, 0], [0, 1], [1, 1]])
        >>> coker = phi.cokernel()
        >>> coker.target.isomorphic(LCA([1, 1, 0]))
        True

        """
        # Horizontally stack A and ker(pi_2)
        A_ker_pi = self.A.row_join(remove_zero_columns(diag(
            *self.target.orders)))
        # Compute SNF, get size and the kernel
        U, S, V = smith_normal_form(A_ker_pi)
        diagonal = nonzero_diag_as_list(S)
        (m, n), r = self.A.shape, len(diagonal)
        quotient = diagonal + [0] * (m - r)

        # Initialize the cokernel morphism and project it onto the target
        coker = type(self)(U, target = quotient, source=self.target)
        return coker.project_to_target()

    def _FGA_image(self):
        """
        Compute the image monomorphism.

        Returns
        -------
        homomorphism : HomLCA
            The image homomorphism.

        Examples
        --------
        >>> phi = HomLCA([[4, 4],
        ...               [2, 8]], target = [64, 32])
        >>> im = phi.image().remove_trivial_groups()
        >>> coim = phi.coimage().remove_trivial_groups()
        >>> phi == (im * coim).project_to_target()
        True

        """
        # Solve equation for the image
        coim = self.coimage()
        coim_target = Matrix(coim.target.orders)
        solved_mat = solve_epi(coim.A, self.A, coim_target)

        # Initialize morphism and return project onto target
        image = type(self)(solved_mat, self.target, source=coim.target)
        return image.project_to_target()

    def _FGA_isomorphic(self, other):
        """
        Whether or not two HomLCAs are isomorphic.

        Two homomorphisms are isomorphic iff they generate the same group.

        Parameters
        ----------
        other : HomLCA
            The homomorphism to compare with.

        Returns
        -------
        isomorphic : bool
            Whether or not `self` and `other` are isomorphic.

        Examples
        --------
        >>> from sympy import diag
        >>> # The order does not matter
        >>> phi = HomLCA(diag(2, 3))
        >>> psi = HomLCA(diag(3, 2))
        >>> phi._FGA_isomorphic(psi)
        True
        >>> # These homomorphisms both generate Z_2 + Z
        >>> phi = HomLCA([[2, 0],
        ...               [0, 1]], target=[4, 0])
        >>> psi = HomLCA([[1, 0],
        ...               [0, 1]], target=[2, 0])
        >>> phi._FGA_isomorphic(psi)
        True
        >>> # They both generate a group isomorphic to Z_6
        >>> phi = HomLCA([[2, 0], [0, 1]], target=[4, 3])
        >>> psi = HomLCA([2, 2], target=[4, 3])
        >>> phi._FGA_isomorphic(psi)
        True
        """
        # TODO: Is this correct?


        self_no_triv = self.remove_trivial_groups().image()
        other_no_triv = other.remove_trivial_groups().image()

        isomorphic = self_no_triv.source.isomorphic(other_no_triv.source)

        return isomorphic

    def _FGA_kernel(self):
        """
        Compute the kernel monomorphism.

        Returns
        -------
        homomorphism : HomLCA
            The kernel homomorphism.

        Examples
        --------
        >>> phi = HomLCA([[1, 0, 1], [0, 1, 1]])
        >>> phi.kernel() == HomLCA([-1, -1, 1])
        True

        """
        # Horizontally stack A and ker(pi_2)
        orders = self.target.orders
        A_ker_pi = self.A.row_join(remove_zero_columns(diag(*orders)))

        # Compute SNF, get size and the kernel
        U, S, V = smith_normal_form(A_ker_pi)
        (m, n), r = self.A.shape, len(nonzero_diag_as_list(S))
        kernel_matrix = V[:n, r:]

        kernel = type(self)(kernel_matrix, target = self.source)
        return kernel.project_to_target()

    def project_to_source(self):
        """
        Project columns to source group (orders).

        Returns
        -------
        homomorphism : HomLCA
            A homomorphism with orders in the source FGA.

        Examples
        --------
        >>> target = [3, 6]
        >>> phi = HomLCA([[1, 0],
        ...               [3, 3]], target = target)
        >>> phi = phi.project_to_source()
        >>> phi.source.orders == [6, 2]
        True
        """
        A_is_integer = all(isinstance(a, self._A_integer_entry_types) for a
                           in self.A)

        if A_is_integer and self.source.is_FGA():
            # Find dimensions
            m, n = self.A.shape

            # Compute orders for all columns of A
            target_vect = Matrix(self.target.orders)
            source = [order_of_vector(self.A[:, i], target_vect) for i in range(n)]
            return type(self)(self.A, self.target, source)
        else:
            raise NotImplementedError('Not implemented: projection to target.')

    def project_to_target(self):
        """
        Project columns to target group.

        Returns
        -------
        homomorphism : HomLCA
            A homomorphism with columns projected to the target FGA.

        Examples
        --------
        >>> target = [7, 12]
        >>> phi = HomLCA([[15, 12],
        ...               [9,  17]], target = target)
        >>> phi_proj = HomLCA([[1, 5],
        ...                    [9, 5]], target = target)
        >>> phi.project_to_target() == phi_proj
        True
        """

        A = matrix_mod_vector(self.A, Matrix(self.target.orders))
        return type(self)(A, target = self.target, source = self.source)

if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose = False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains factorization algorithms for matrices over the
integers. All the inputs and outputs are of type
:py:class:`~sympy.matrices.dense.MutableDenseMatrix`.
"""


from sympy import Matrix, gcdex
from abelian.utils import mod

def hermite_normal_form(A):
    """
    Compute U and H such that A*U = H.

    This algorithm computes the column version of the Hermite normal form, and
    returns a tuple of matrices (U, H) such that A*U = H. The matrix U is an
    unimodular transformation matrix and H is the result of the transformation,
    i.e. H is in Hermite normal form.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The matrix to decompose.

    Returns
    -------
    U : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        An unimodular matrix, i.e. integer matrix with determinant +/- 1.

    H : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A matrix in Hermite normal form.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[1, 2],
    ...             [3, 4]])
    >>> U, H = hermite_normal_form(A)
    >>> # Verify that U is unimodular (determinant +/- 1)
    >>> U.det() in [1, -1]
    True
    >>> # Verify the decomposition
    >>> A*U == H
    True
    """

    # Get size and set up matrices U and H
    m, n = A.shape
    j = 0
    H = A.copy()
    U = Matrix.eye(n)

    # Iterate down the rows of the matrix
    for i in range(0, m):

        # If every entry to the right is a zero, no pivot will be found
        # for this row and we move on to the next one.
        if H[i, j:] == H[i, j:] * 0:
            continue

        # Create zeros to the right of the pivot H[i, j]
        for k in range(j + 1, n):

            # Skip the column if the element is zero
            # In this case the column index j in not incremented
            if H[i, k] == 0:
                continue

            # Apply the 'elementary hermite transform' to the columns of H,
            # ignoring the top i rows of H as they are identically zero
            # The implementation of the 'elementary hermite transform'
            # does not explicitly compute matrix products.
            # Equivalent to right-multiplication by Matrix([[a, -s/g],
            #                                               [b, r/g]])

            # Extended Euclidean algorithm, so that r*a + s*b = g = gcd(r, s)
            r, s = H[i, j], H[i, k]
            a, b, g = gcdex(r, s)

            # Apply the matrix product of H and U
            H[i:, j], H[i:, k] = (a * H[i:, j] + b * H[i:, k],
                                  -(s / g) * H[i:, j] + (r / g) * H[i:, k])
            U[:, j], U[:, k] = (a * U[:, j] + b * U[:, k],
                                -(s / g) * U[:, j] + (r / g) * U[:, k])

        # Make sure the pivot element is positive.
        # Some saving achieved by realizing that the first i rows of H are
        # identically zero -- thus no multiplication is needed.
        if H[i, j] < 0:
            H[i:, j] = -H[i:, j]
            U[:, j] = -U[:, j]

        # Making all elements to the left of the pivot H[i, j] smaller
        # than the pivot and positive using division algorithm transform
        for k in range(0, j):
            # Compute quotient in the division algorithm, subtracting
            # the quotient times H[:, j] leaves a positive remainder
            a = H[i, k] // H[i, j]
            H[:, k] = H[:, k] - a * H[:, j]
            U[:, k] = U[:, k] - a * U[:, j]
        # Increment j (the column index). Break if j is out of dimension
        j += 1
        if j >= n:
            # j is out of dimension, break
            break

    return U, H

def smith_normal_form(A, compute_unimod = True):
    """
    Compute U,S,V such that U*A*V = S.

    This algorithm computes the Smith normal form of an integer matrix.
    If `compute_unimod` is True, it returns matrices (U, S, V) such
    that U*A*V = S, where U and V are unimodular and S is in Smith normal
    form. If `compute_unimod` is False, it returns S and does not compute U
    and V.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The matrix to factor.
    compute_unimod : bool
        Whether or not to compute and return unimodular matrices U and V.

    Returns
    -------
    U : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        An unimodular matrix, i.e. integer matrix with determinant +/- 1.

    S : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A matrix in Smith normal form.

    V : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        An unimodular matrix, i.e. integer matrix with determinant +/- 1.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[1, 2],
    ...             [3, 4]])
    >>> U, S, V = smith_normal_form(A)
    >>> # Verify that U and V are both unimodular
    >>> U.det() in [1, -1] and V.det() in [1, -1]
    True
    >>> # Verify the factorization
    >>> U * A * V == S
    True
    >>> # Compute without U and V, verify that the result is the same
    >>> K = smith_normal_form(A, compute_unimod=False)
    >>> K == S
    True
    """

    # Get size and set up the unimodular matrices U and V
    m, n = A.shape
    min_m_n = min(m, n)
    S = A.copy()
    if compute_unimod:
        U, V = Matrix.eye(m), Matrix.eye(n)

    def row_col_all_zero(matrix, f):
        """
        Check that all entries to the right of and below `f` are zero.
        """
        for entry in matrix[f, f + 1:]:
            if entry != 0:
                return False
        for entry in matrix[f + 1:, f]:
            if entry != 0:
                return False
        return True

    # Main loop, iterate over all sub-matrices to reduce
    f = 0
    while f < min_m_n:

        # While there are non-zero elements to reduce in row/column f
        # and the diagonal element is not positive
        while not (row_col_all_zero(S, f) and S[f, f] >= 0):

            # Find index pair of minimum non-zero entry (in absolute value)
            # in the sub-matrix S[f:, f:].
            indices = ((i, j) for j in range(f, n) for i in range(f, m))
            key_val_pairs = ((index, abs(S[index])) for index in indices
                             if abs(S[index]) != 0)
            (i, j), min_val = min(key_val_pairs, key=lambda k: k[1])

            # Permute S to move the minimal element to the pivot location
            S[f:, j], S[f:, f] = S[f:, f], S[f:, j]
            S[i, f:], S[f, f:] = S[f, f:], S[i, f:]
            if compute_unimod:
                V[:, j], V[:, f] = V[:, f], V[:, j]
                U[i, :], U[f, :] = U[f, :], U[i, :]

            # If the freshly permuted pivot is negative, make it positive
            if S[f, f] < 0:
                S[f:, f] = -S[f:, f]
                if compute_unimod:
                    V[:, f] = -V[:, f]

            # Reduce row f so every entry is smaller than pivot
            for k in range(f + 1, n):
                if S[f, k] == 0:
                    continue

                # Subtract a times column f from column k
                a = S[f, k] // S[f, f]
                S[f:, k] = S[f:, k] - a * S[f:, f]
                if compute_unimod:
                    V[:, k] = V[:, k] - a * V[:, f]

            # Reduce column f so every entry is smaller than pivot
            for k in range(f + 1, m):
                if S[k, f] == 0:
                    continue

                # Subtract a times row f from row k
                a = S[k, f] // S[f, f]
                S[k, f:] = S[k, f:] - a * S[f, f:]
                if compute_unimod:
                    U[k, :] = U[k, :] - a * U[f, :]

        f += 1

    # Enforce divisibility criterion using the 'divisibility transformation'
    # matrices.
    for f in range(min_m_n):
        for k in range(f + 1, min_m_n):

            # Divisibility criterion is fulfilled
            if mod(S[k, k], S[f, f]) == 0:
                continue

            # S[f, f] does not divide S[k, k]
            r, s = S[f, f], S[k, k]
            a, b, c = gcdex(r, s)
            S[f, f], S[k, k] = c, (r * s) // c
            if compute_unimod:
                V[:, f], V[:, k] = (V[:, f] + V[:, k],
                                    -b * (s / c) * V[:,f] + a * (r / c) * V[:, k])
                U[f, :], U[k, :] = (a * U[f, :] + b * U[k, :],
                                    -(s / c) * U[f, :] + (r / c) * U[k, :])

    if compute_unimod:
        return U, S, V
    else:
        return S


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose = True)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
from abelian import LCA, LCAFunc, HomLCA, voronoi
from sympy import Matrix, diag, Integer, Float, Rational


def close(a, b):
    numeric_types = (float, int, complex, Integer, Float, Rational)
    if isinstance(a, numeric_types) and isinstance(a, numeric_types):
        return abs(a-b) < 10e-10
    return sum(abs(i-j) for (i, j) in zip(a, b))

class TestThesisExamples:

    def test_example_1_HomFGA(self):
        """
        Test example 1 from the abelian introductory chapter
        in the thesis.
        """

        # Import the classes and create a homomorphism
        from abelian import HomLCA, LCA
        target = LCA([8, 5])  # Create Z_8 + Z_5
        phi = HomLCA([[4, 2], [7, 3]], target=target)

        # Compute cokernel, then remove trivial groups
        cokernel = phi.cokernel().remove_trivial_groups()

        # Compute image, then remove trivial groups
        image = phi.image().remove_trivial_groups()

        # Compute coimage, remove trivial, then project
        coimage = phi.coimage().remove_trivial_groups()
        coimage = coimage.project_to_source()

        # Project phi, compute kernel
        phi_projected = phi.project_to_source()
        kernel = phi_projected.kernel().project_to_source()

        assert cokernel.A == Matrix([[1, 0]])
        assert image.A == Matrix([[2], [3]])
        assert coimage.A == Matrix([[14, 1]])
        assert kernel.A == Matrix([[2, 5], [12, 10]])

    def test_example_2_FourierSeries(self):
        """
        Test example 2 from the abelian introductory chapter
        in the thesis.
        """
        # Import objects, create function on T^2
        from abelian import HomLCA, LCA, LCAFunc
        from sympy import Rational, diag
        T = LCA(orders=[1], discrete=[False])
        func = LCAFunc(lambda x: sum(x), domain=T ** 2)

        # Create homomorphism to sample function
        n = 10
        Z_n = LCA(orders=[n], discrete=[True])
        phi = HomLCA(diag(Rational(1, n), Rational(1, n)),
                     target=T ** 2, source=Z_n ** 2)

        # Sample, dualize
        func_sampled = func.pullback(phi)
        func_sample_dual = func_sampled.dft()

        # Transversal - minimizes distance
        func_dual = func_sample_dual.transversal(phi.dual())

        assert func([0.5, 0.5]) == 1
        assert close(func([1.6, 0.6]), 1.2)
        assert close(func_sampled([11, 11]), func([0.1, 0.1]))
        assert close(func_sampled([1, 1]),   func([0.1, 0.1]))

        assert func_dual([11, 11]) == 0
        assert close(func_dual([0, 0]), 0.9)

    def test_example_3_Hexagonal(self):
        from math import exp, sqrt, pi

        # Import objects, create function on R^n
        from abelian import HomLCA, LCA, LCAFunc, voronoi
        R = LCA(orders=[0], discrete=[False])
        k = 0.5  # Decay of exponential
        func_expr = lambda x: exp(-k * sum(x_j ** 2 for x_j in x))
        func = LCAFunc(func_expr, domain=R ** 2)

        # Create a homomorphism to sample
        hexagonal_generators = [[1, 0.5], [0, sqrt(3) / 2]]
        phi_sample = HomLCA(hexagonal_generators, target=R ** 2)

        # Create a homomorphism to periodize
        n = 3
        phi_periodize = HomLCA([[n, 0], [0, n]])
        coker_phi_p = phi_periodize.cokernel()

        # Move function from R**2 to Z**2 to Z_n**2
        func_sampled = func.pullback(phi_sample)
        func_periodized = func_sampled.pushforward(coker_phi_p, 25)

        # Move function to dual space, then to T**2
        func_dual = func_periodized.dft()
        phi_periodize_ann = phi_periodize.annihilator()

        # Compute a Voronoi transversal function, interpret on R**2
        sigma = voronoi(phi_sample.dual(), norm_p=2)
        for element in func_dual.domain.elements_by_maxnorm():
            value = func_dual(element)
            coords_on_R = sigma(phi_periodize_ann(element))

        approx, true = func_dual([0, 0])*n*n, sqrt((2*pi)**2)
        print(approx, true, true/approx)

        print(phi_sample.A.det()*2)

        print(approx*phi_sample.A.det()*2, true/k)


if __name__ == '__main__':
    t = TestThesisExamples()
    t.test_example_3_Hexagonal()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module consists of a class for elementary locally compact abelian groups,
the LCA class.
"""

from sympy import Matrix, Integer, diag
from abelian.utils import mod
from abelian.linalg.factorizations import smith_normal_form
from abelian.linalg.utils import nonzero_diag_as_list
from collections.abc import Sequence, Callable
from abelian.linalg.free_to_free import elements_of_maxnorm_FGA
import itertools

class LCA(Sequence, Callable):
    """
    An elementary locally compact abelian group (LCA).
    """

    # These dictionaries lets the user initialize using
    _repr_dict = {True: 'd', False :'c'}
    _repr_dict_inv = {'d':True, 'c':False}

    _integer_types = (int, Integer)

    def __init__(self, orders, discrete = None):
        """
        Initialize a new LCA.

        This class represents locally compact abelian groups, defined by their
        orders and whether or not they are discrete. An order of 0 means
        infinite order. The possible groups are:

        * :math:`\mathbb{Z}_n` : order = `n`, discrete = `True`
        * :math:`\mathbb{Z}` : order = `0`, discrete = `True`
        * :math:`T` : order = `1`, discrete = `False`
        * :math:`\mathbb{R}` : order = `0`, discrete = `False`

        Every locally compact abelian group is isomorphic to a direct sum
        or one or several of the groups above.

        Parameters
        ----------
        orders : list
            A list of orders, e.g. [6, 8, 11].
        discrete : list
            A list of booleans such as [True, False, ...] or alternatively a
            list of letters such as ['d', 'c', ...], where 'd' stands for
            discrete and 'c' stands for continuous. If None, it defaults to
            discrete.


        Examples
        ---------
        >>> # Create G = Z_5 + Z_6 + Z_7 in three ways
        >>> G1 = LCA([5, 6, 7])
        >>> G2 = LCA([5, 6, 7], [True, True, True])
        >>> G3 = LCA([5, 6, 7], ['d']*3)
        >>> (G1 == G2 == G3)
        True

        >>> # Create G = R + Z
        >>> G = LCA(orders = [0, 0], discrete = [False, True])

        >>> G = LCA([], [])
        >>> G
        []
        """

        orders, discrete = self._verify_init(orders, discrete)
        self.orders = orders
        self.discrete = discrete


    @classmethod
    def trivial(cls):
        """
        Return a trivial LCA.

        Returns
        --------
        group : LCA
            A trivial LCA.

        Examples
        ---------
        >>> trivial = LCA.trivial()
        >>> Z = LCA([0])
        >>> (Z + trivial).isomorphic(Z)
        True
        """
        return cls(orders = [1], discrete = [True])


    @classmethod
    def _verify_init(cls, orders, discrete):
        """
        Verify the user inputs.

        Return `orders` and `discrete` as lists.
        """

        # Forgive if integer types have been passed, or if it's a matrix
        if isinstance(orders, (Matrix,) + cls._integer_types):
            orders = list(orders)

        # Forgive if boolean True/False og string 'd'/'c' was passed
        if isinstance(discrete, (Matrix, bool, str)):
            discrete = list(discrete)

        # From here on out, assume `orders` and `discrete` are lists
        if len(orders) < 0:
            raise ValueError('List of orders must have length >=0.')

        if any(i < 0 for i in orders):
            raise ValueError('Every order must be >= 0.')

        if discrete is None:
            discrete = [True] * len(orders)

        if len(orders) != len(discrete):
            raise ValueError('Orders and list of discrete must match length.')

        for p, d in zip(orders, discrete):
            if (not d) and (p not in [0, 1]):
                raise ValueError('Continuous groups must have order 0 or 1.')

        def map_to_bool(x):
            try:
                return cls._repr_dict_inv[x.lower()]
            except:
                return x

        discrete = [map_to_bool(i) for i in discrete]

        return orders, discrete


    def __add__(self, other):
        """
        Override the addition (`+`) operator,
        see :py:meth:`~abelian.groups.LCA.sum`.
        """
        return self.sum(other)

    def __call__(self, element):
        """
        Override function calls,
        see :py:meth:`~abelian.groups.LCA.project_element`.
        """
        return self.project_element(element)

    def __contains__(self, other):
        """
        Override the 'in' operator,
        see :py:meth:`~abelian.groups.LCA.contained_in`.
        """
        return other.contained_in(self)

    def __eq__(self, other):
        """
        Override the equality (`==`) operator,
        see :py:meth:`~abelian.groups.LCA.equal`.
        """
        return self.equal(other)

    def __getitem__(self, key):
        """
        Override the slice operator,
        see :py:meth:`~abelian.groups.LCA.slice`.
        """
        return self.getitem(key)

    def __iter__(self):
        """
        Override the iteration protocol,
        see :py:meth:`~abelian.groups.LCA.iterate`.
        """
        yield from self.iterate()

    def __len__(self):
        """
        Override the ``len()`` function,
        see :py:meth:`~abelian.groups.LCA.length`.
        """
        return self.length()

    def __pow__(self, power, modulo=None):
        """
        Override the pow (`**`) operator,
        see :py:meth:`~abelian.groups.LCA.compose_self`.
        """
        return self.compose_self(power)

    def __repr__(self):
        """
        Override the ``repr()`` function.

        This method is called for computer representation, and if no
        __str__ method is defined this function will be called when ``print()``
        is called on an instance too.
        """

        def repr_single(p, d):
            if p == 0:
                if d:
                    return r'Z'
                return r'R'
            if d:
                return r'Z_' + str(p) + ''
            return r'T' # r'T_' + str(p) + ''

        tuples = self._iterate_tuples()
        joined = r', '.join(repr_single(p, d) for (p, d) in tuples)
        return '[' + joined + ']'

    def canonical(self):
        """
        Return the LCA in canonical form using SNF.

        The canonical form decomposition will:

        (1) Put the torsion (discrete with order >= 1) subgroup in
            a canonical form using invariant factor decomposition from
            the Smith Normal Form decomposition.
        (2) Sort the non-torsion subgroup.

        Returns
        --------
        group : LCA
            The LCA in canonical form.

        Examples
        ---------
        >>> G = LCA([4, 3])
        >>> G.canonical() == LCA([12])
        True
        >>> G = LCA([1, 1, 8, 2, 4], ['c', 'd', 'd', 'd', 'd'])
        >>> G.canonical() == LCA([1], ['c']) + LCA([2, 4, 8])
        True
        """

        def is_t(order, discrete):
            """
            Function to determine if a group is torsion or not.
            """
            if order > 0 and discrete:
                return True
            return False

        def split_torsion(LCA):
            """
            Split a group into (torsion, non_torsion).
            """
            gen_list = list(enumerate(LCA._iterate_tuples()))
            torsion_i = [i for (i, (p, d)) in gen_list if is_t(p, d)]
            nontorsion_i = [i for (i, (p, d)) in gen_list if not is_t(p, d)]
            torsion = LCA.remove_indices(nontorsion_i)
            non_torsion = LCA.remove_indices(torsion_i)
            return torsion, non_torsion

        # Get information pertaining to self
        self_tors, self_non_tors = split_torsion(self)

        # Sort the non-torsion subgroup
        tuples = list(self_non_tors._iterate_tuples())
        non_tors_orders = [p for (p, d) in sorted(tuples)]
        non_tors_discrete = [d for (p, d) in sorted(tuples)]

        # Get canonical decomposition of the torsion subgroup
        self_SNF = smith_normal_form(diag(*self_tors.orders), False)
        self_SNF_p = [p for p in nonzero_diag_as_list(self_SNF) if p != 1]

        # Construct the new group
        orders = non_tors_orders + self_SNF_p
        discrete = non_tors_discrete + [True] * len(self_SNF_p)
        return type(self)(orders = orders, discrete = discrete)

    def copy(self):
        """
        Return a copy of the LCA.

        Returns
        --------
        group : LCA
            A copy of the LCA.

        Examples
        ---------
        >>> G = LCA([1, 5, 7], [False, True, True])
        >>> H = G.copy()
        >>> G == H
        True
        """
        orders_cp = self.orders.copy()
        discrete_cp = self.discrete.copy()
        return type(self)(orders = orders_cp, discrete = discrete_cp)

    def dual(self):
        """
        Return the Pontryagin dual of the LCA.

        Returns a group isomorphic to the Pontryagin dual.

        Returns
        -------
        group : LCA
            The Pontryagin dual of the LCA.

        Examples
        ---------
        >>> G = LCA([5, 1], [True, False])
        >>> H = LCA([5, 0], [True, True])
        >>> G.dual() == H
        True
        >>> G.dual().dual() == G
        True
        >>> self_dual = LCA([5])
        >>> self_dual.dual() == self_dual
        True
        """

        single_dual = self._dual_of_single_group
        dual_lists = list(single_dual(p, d) for (p, d) in self._iterate_tuples())
        new_orders = [p for (p, d) in dual_lists]
        new_discrete = [d for (p, d) in dual_lists]
        return type(self)(orders = new_orders, discrete = new_discrete)

    def elements_by_maxnorm(self, norm_values = None):
        """
        Yield elements corresponding to max norm value.

        If the group is discrete, elements can be generated by maxnorm.

        Parameters
        ----------
        norm_values : iterable
            An iterable containing integer norm values.

        Yields
        -------
        group_element : list
            Group elements with max norm specified by the input iterable.

        Examples
        ---------
        >>> G = LCA([0, 0])
        >>> for element in G.elements_by_maxnorm([0, 1]):
        ...     print(element)
        [0, 0]
        [1, -1]
        [-1, -1]
        [1, 0]
        [-1, 0]
        [1, 1]
        [-1, 1]
        [0, 1]
        [0, -1]
        >>> G = LCA([5, 8])
        >>> for element in G.elements_by_maxnorm([4, 5]):
        ...     print(element)
        [3, 4]
        [4, 4]
        [0, 4]
        [1, 4]
        [2, 4]
        """
        if not all(self.discrete):
            raise TypeError('The group must be discrete.')


        # If the argument is none, create a counter
        if not norm_values:
            norm_values = itertools.count(start = 0, step = 1)
        if isinstance(norm_values, int):
            norm_values = [norm_values]

        orders = self.orders
        for maxnorm_value in norm_values:

            # Keep track of whether ot not a value is yielded
            yielded = False
            for element in elements_of_maxnorm_FGA(orders, maxnorm_value):
                yielded = True
                yield list(element)

            # If no value is yielded, break the possibly infinite loop
            if not yielded:
                break

    def equal(self, other):
        """
        Whether or not two LCAs are equal.

        Two LCAs are equal iff the list of `orders` and the list of `discrete`
        are both equal.

        Parameters
        ----------
        other : LCA
            The LCA to compare equality with.

        Returns
        --------
        equal : bool
            Whether or not the LCAs are equal.

        Examples
        ---------
        >>> G = LCA([1, 5, 7], [False, True, True])
        >>> H = G.copy()
        >>> G == H  # The `==` operator is overloaded
        True
        >>> G.equal(H)  # Equality using the method
        True
        """
        orders_equal = self.orders == other.orders
        discrete_equal = self.discrete ==  other.discrete
        return (orders_equal and discrete_equal)


    def getitem(self, key):
        """
        Return a slice of the LCA.

        Parameters
        ----------
        key: :py:class:`~python.slice`
            A slice object, or an integer.

        Returns
        -------
        group: LCA
            A slice of the FGA as specified by the slice object.

        Examples
        ---------
        >>> G = LCA([5, 6, 1])
        >>> G[0:2] == LCA([5, 6])
        True
        >>> G[0] == LCA([5])
        True
        """

        orders = self.orders[key]
        discrete = self.discrete[key]
        orders = orders if isinstance(orders, list) else [orders]
        discrete = discrete if isinstance(discrete, list) else [discrete]
        return type(self)(orders = orders, discrete = discrete)


    def is_FGA(self):
        """
        Whether or not the LCA is a FGA.

        A locally compact abelian group (LCA) is a finitely generated
        abelian group (FGA) iff all the groups in the direct sum are discrete.

        Returns
        -------
        is_FGA : bool
            True if the object is an FGA, False if not.

        Examples
        ---------
        >>> G = LCA([5, 1], [True, False])
        >>> G.is_FGA()
        False
        >>> G = LCA([1, 7], [False, True])
        >>> G.dual().is_FGA()
        True
        """
        is_FGA = all(self.discrete)
        return is_FGA

    def isomorphic(self, other):
        """
        Whether or not two LCAs are isomorphic.

        Two LCAs are isomorphic iff they can be put into the same
        canonical form.

        Parameters
        ----------
        other : LCA
            The LCA to compare with.

        Returns
        --------
        isomorphic : bool
            Whether or not the LCAs are isomorphic.

        Examples
        ---------
        >>> G = LCA([3, 4])
        >>> H = LCA([12])
        >>> G.isomorphic(H)
        True
        >>> G.equal(H)
        False

        >>> LCA([2, 6]).isomorphic(LCA([12]))
        False

        >>> G = LCA([0, 0, 1, 3, 4], [False, True, True, True, True])
        >>> H = LCA([0, 0, 3, 4], [True, False, True, True])
        >>> G.isomorphic(H)
        True

        >>> LCA([]).isomorphic(LCA.trivial())
        True
        """
        isomorphic = self.canonical().equal(other.canonical())
        return isomorphic

    def iterate(self):
        """
        Yields the groups in the direct sum one by one.

        Iterate through the groups and yield the individual groups in the
        direct sum, one by one.

        Yields
        -------
        group : LCA
            A single group in the direct sum.

        Examples
        ---------
        >>> G = LCA([5, 1], [True, False])
        >>> groups = [LCA([5], [True]), LCA([1], [False])]
        >>> for i, group in enumerate(G):
        ...     group == groups[i]
        True
        True
        """
        for (order, discrete) in zip(self.orders, self.discrete):
            yield type(self)([order], [discrete])

    def length(self):
        """
        The number of groups in the direct sum.

        Returns
        -------
        length : int
            The number of groups in the direct sum.

        Examples
        --------
        >>> G = LCA([])
        >>> G.length()
        0

        >>> G = LCA([0, 1, 1, 5])
        >>> G.length()
        4
        """
        return len(self.orders)

    def compose_self(self, power):
        """
        Repeated direct summation.

        Returns
        -------
        group : LCA
            A new group.

        Examples
        --------
        >>> R = LCA([0], [False])
        >>> (R + R) == R**2
        True
        >>> R**0 == LCA.trivial()
        True
        >>> Z = LCA([0])
        >>> (Z + R)**2 == Z + R + Z + R
        True

        """
        if not isinstance(power, int):
            raise ValueError('Power must be an integer.')
        if power <= 0:
            return self.trivial()
        if power == 1:
            return self
        if power > 1:
            direct_sum = self.copy()
            for prod in range(power - 1):
                direct_sum = direct_sum.sum(self)
            return direct_sum

    def project_element(self, element):
        """
        Project an element onto the group.

        Parameters
        -----------
        element : :py:class:`~sympy.matrices.dense.MutableDenseMatrix` or list
            The group element to project to the LCA.

        Returns
        -------
        element : :py:class:`~sympy.matrices.dense.MutableDenseMatrix` or list
            The group element projected to the LCA.

        Examples
        ---------
        >>> from sympy import Matrix
        >>> G = LCA([5, 9])
        >>> g = [6, 22]
        >>> G.project_element(g)
        [1, 4]
        >>> g = Matrix([13, 13])
        >>> G.project_element(g) == Matrix([3, 4])
        True
        """
        if len(element) in [0, 1] and self.isomorphic(self.trivial()):
            return [0]

        if not (len(element) == len(self.orders) == len(self.discrete)):
            raise ValueError('Length of element must match groups.')

        def project(element, order, discrete):
            if not discrete:
                return mod(element, order)
            if discrete:
                #if mod(element, 1) == 0:
                if isinstance(element, self._integer_types):
                    return mod(element, order)

                raise ValueError('Non-integer cannot be projected to '
                                 'discrete group.')

        generator = zip(element, self.orders, self.discrete)
        projected = [project(e, p, d) for (e, p, d) in generator]

        # If the input is a list, return a list
        if isinstance(element, list):
            return projected

        # If the input is a sympy Matrix, return a sympy Matrix
        if isinstance(element, Matrix):
            return Matrix(projected)


    def rank(self):
        """
        Return the rank of the LCA.

        Returns
        --------
        rank : int
            An integer greater than or equal to 0.

        Examples
        ---------
        >>> G = LCA([5, 6, 1])
        >>> G.rank()
        2

        >>> LCA([1]).rank()
        0

        >>> G = LCA([5, 6, 1])
        >>> H = LCA([1])
        >>> G.rank() + H.rank() == (G + H).rank()
        True
        """

        removed_trivial = self.remove_trivial()
        return removed_trivial.length()

    def remove_indices(self, indices):
        """
        Return a LCA with some groups removed.

        Parameters
        ----------
        indices : list
            A list of indices corresponding to LCAs to remove.

        Returns
        -------
        group : LCA
            The LCA with some groups removed.

        Examples
        --------
        >>> G = LCA([5, 8, 9])
        >>> G.remove_indices([0, 2]) == LCA([8])
        True

        """
        enu = enumerate
        orders = [p for (i, p) in enu(self.orders) if i not in indices]
        discrete = [p for (i, p) in enu(self.discrete) if i not in indices]
        return type(self)(orders = orders, discrete = discrete)

    def remove_trivial(self):
        """
        Remove trivial groups from the object.

        Returns
        --------
        group : LCA
            The group with trivial groups removed.

        Examples
        ---------
        >>> G = LCA([5, 1, 1])
        >>> G.remove_trivial() == LCA([5])
        True
        """
        def trivial(order, discrete):
            return discrete and (order == 1)

        self_tuples = self._iterate_tuples()
        purged_lists = [(p, d) for (p, d) in self_tuples if not trivial(p, d)]
        new_orders = [p for (p, d) in purged_lists]
        new_discrete = [d for (p, d) in purged_lists]
        return type(self)(orders=new_orders, discrete=new_discrete)

    def contained_in(self, other):
        """
        Whether the LCA is contained in `other`.

        A LCA G is contained in another LCA H iff there exists an injection
        from the elements of G to H such that every source/target of the
        mapping is isomorphic. In other words, every group in G must be found
        in H, and no two groups in G can be identified with the same isomorphic
        group is H.

        Parameters
        ----------
        other : LCA
            A locally compact abelian group.

        Returns
        -------
        is_subgroup : bool
            Whether or not `self` is contained in other.

        Examples
        --------
        >>> # Simple example
        >>> G = LCA([2, 2, 3])
        >>> H = LCA([2, 2, 3, 3])
        >>> G.contained_in(H)
        True
        >>> # Order does not matter
        >>> G = LCA([2, 3, 2])
        >>> H = LCA([2, 2, 3, 3])
        >>> G.contained_in(H)
        True
        >>> # Trivial groups are not removed
        >>> G = LCA([2, 3, 2, 1])
        >>> H = LCA([2, 2, 3, 3])
        >>> G in H
        False

        """
        if not isinstance(other, type(self)):
            return TypeError('Must be LCAs.')

        self_as_list = list(self._iterate_tuples())
        other_as_list = list(other._iterate_tuples())

        for group in self_as_list:
            # If the group is not found, return False
            if group not in other_as_list:
                return False

            # If the group is found, remove it
            other_as_list.remove(group)

        # Every group in G is found in H, and it's a subgroup
        return True

    def sum(self, other):
        """
        Return the direct sum of two LCAs.

        Parameters
        ----------
        other : LCA
            The LCA to take direct sum with.

        Returns
        --------
        group : LCA
            The direct sum of self and other.


        Examples
        ---------
        >>> G = LCA([5])
        >>> H = LCA([7])
        >>> G + H == LCA([5, 7])  # The `+` operator is overloaded
        True
        >>> G + H == G.sum(H)  # Directs sums two ways
        True
        """
        new_orders = self.orders + other.orders
        new_discrete = self.discrete + other.discrete
        return type(self)(orders=new_orders, discrete=new_discrete)

    def to_latex(self):
        """
        Return the LCA as a :math:`\LaTeX` string.

        Returns
        -------
        latex_str : str
            A string with LaTeX code for the object.

        Examples
        ---------
        >>> G = LCA([5, 0], [True, False])
        >>> G.to_latex()
        '\\\mathbb{Z}_{5} \\\oplus \\\mathbb{R}'
        """
        def repr_single(p, d):
            if p == 0:
                if d:
                    return r'\mathbb{Z}'
                return r'\mathbb{R}'
            if d:
                return r'\mathbb{Z}_{' + str(p) + '}'
            return r'T' #r'T_{' + str(p) + '}'

        tuples = self._iterate_tuples()
        return r' \oplus '.join(repr_single(p, d) for (p, d) in tuples)

    def _iterate_tuples(self):
        """
        Yields tuples in the form (`order`, `discrete`).
        """
        for group in self:
            yield (group.orders[0], group.discrete[0])

    @staticmethod
    def _dual_of_single_group(order, discrete):
        """
        Compute the dual of a single group.
        """
        if discrete:
        # Discrete
            if order == 0:
                # Dual of Z is T
                return [1, False]  # Return T
            else:
                # Dual of Z_n is Z_n
                return [order, True]  # Return Z_n

        else:
        # Continuous
            if order == 0:
                # Dual of R is R
                return [0, False]  # Return R
            else:
                # Dual of T is Z
                return [0, True]  # Return Z



if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose = False)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
from sympy import Matrix
from random import randint as ri
from random import random
from abelian.morphisms import HomLCA, LCA


def frob_norm(A, B):
    """
    Frobenius norm.
    """
    return sum(abs(i-j) for (i, j) in zip(A, B))


class TestSNF:

    @classmethod
    def setup_class(cls):
        m, n = ri(2, 4), ri(2, 4)
        A = Matrix(m, n, lambda i, j: random())

        # Add an extra column to A
        new_col = A[:, 0] + A[:, 1]
        A = A.row_join(new_col)

        # Add an extra row to A
        new_row = A[0, :] + A[1, :]
        A = A.col_join(new_row)

        R = LCA([0], [False])
        cls.phi = HomLCA(A, source = R**(n+1), target = R**(m+1))


    def test_real_kernel(self):
        """
        Test the real kernel.
        """
        phi = self.phi
        phi_ker = phi.kernel()

        A = (phi * phi_ker).A
        B = HomLCA.zero(target = phi.target, source =phi_ker.source).A
        assert frob_norm(A, B) < 10e-10


    def test_real_cokernel(self):
        """
        Test the real cokernel.
        """
        phi = self.phi
        phi_coker = phi.cokernel()

        A = (phi_coker * phi).A
        B = HomLCA.zero(target = phi_coker.target, source = phi.source).A
        assert frob_norm(A, B) < 10e-10


    def test_real_im_coim(self):
        """
        Test the real image/coimage.
        """
        phi = self.phi
        phi_image = phi.image()
        phi_coimage = phi.coimage()

        B = (phi_image * phi_coimage).A
        assert frob_norm(phi.A, B) < 10e-10





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains functions which calculate mapping properties of
free-to-free homomorphisms. All the inputs and outputs are of type
:py:class:`~sympy.matrices.dense.MutableDenseMatrix`.
"""

import itertools
from sympy import Matrix, Integer, Float, Rational
from abelian.linalg.factorizations import smith_normal_form
from abelian.linalg.utils import nonzero_diag_as_list
from collections.abc import Iterable



def mod(a, b):
    """
    Mod for integers, tuples and lists.

    Parameters
    ----------
    a : int, tuple or list
        The argument.
    b : int, tuple or list
        The order.

    Returns
    -------
    int, tuple or list
        A mod b.

    Examples
    ---------
    >>> mod(7, 5) # Integer data
    2
    >>> mod((5, 8), (4, 4)) # Tuple data
    (1, 0)
    >>> mod([5, 8], [4, 4]) # List data
    [1, 0]
    """
    integer_types = (int, Integer, float, Float, Rational)

    # It's a numeric, non-iterable data type
    if isinstance(a, integer_types) and isinstance(b, integer_types):
        if b == 0:
            return a
        return a % b

    # It's an iterable data type
    if isinstance(a, Iterable) and isinstance(b, Iterable):
        return type(a)([mod(i, j) for i, j in zip(a, b)])

    raise TypeError('Did not recognize data type for modulus.')

def elements_increasing_norm(free_rank, end_value = None):
    """
    Continually yield every element in Z^r of increasing max-norm.

    Parameters
    ----------
    free_rank : int
        The free rank (like dimension) of Z^r, i.e. free_rank = r.

    Yields
    -------
    tuple
        Elements in Z^r with increasing maxnorm.

    Examples
    ---------
    >>> free_rank = 2 # Like dimension
    >>> for count, element in enumerate(elements_increasing_norm(free_rank)):
    ...     if count >= 9:
    ...         break
    ...     print(count, element, max(abs(k) for k in element))
    0 (0, 0) 0
    1 (1, -1) 1
    2 (-1, -1) 1
    3 (1, 0) 1
    4 (-1, 0) 1
    5 (1, 1) 1
    6 (-1, 1) 1
    7 (0, 1) 1
    8 (0, -1) 1
    """
    for maxnorm_value in itertools.count(start = 0, step = 1):
        if end_value is not None:
            if maxnorm_value == end_value:
                break
        yield from elements_of_maxnorm(free_rank, maxnorm_value)

def elements_of_maxnorm_FGA(orders, maxnorm_value):
    """
    Yield every element of Z_`orders` such that max_norm(element) = maxnorm_value.

    Parameters
    ----------
    orders : list
        Orders in Z_orders, where 0 means infinite order,
        i.e. [2, 0] is Z_2 + Z.
    maxnorm_value : int
        The value of the maximum norm of the elements generated.

    Yields
    -------
    tuple
        Elements in Z_orders that satisfy the norm criterion.

    Examples
    ---------
    >>> orders = [0, 0]
    >>> norm_value = 1
    >>> elements = list(elements_of_maxnorm_FGA(orders, norm_value))
    >>> len(elements)
    8
    >>> orders = [0, 3]
    >>> norm_value = 2
    >>> for element in elements_of_maxnorm_FGA(orders, norm_value):
    ...     print(element)
    (2, 2)
    (-2, 2)
    (2, 0)
    (-2, 0)
    (2, 1)
    (-2, 1)
    """

    # The zeroth layer is just (0,0,...), yield this and return to terminate
    if maxnorm_value == 0:
        yield tuple([0] * len(orders))
        return

    # Will be used in the loop, so we compute it out-of-loop here
    dimension = len(orders)

    # The 'wall' is the dimension held constant
    for wallnum, dim in enumerate(orders):

        # If the wall is outside the dimension, skip it
        if (dim != 0) and maxnorm_value > (dim // 2):
            continue

        # Set up the cartesian product argument, making sure to remove
        # boundary elements so they are not yielded twice
        boundary_reduced = [1] * wallnum + [0] * (dimension - wallnum - 1)
        prod_arg = [range(-maxnorm_value + k, maxnorm_value + 1 - k) \
                    for k in boundary_reduced]

        # The dimensions that are not constant
        non_const_dims = orders[:]
        non_const_dims.pop(wallnum)
        # non_const_dims = dimensions[:wallnum] + dimensions[wallnum+1:]

        # Go through every argument in the cartesian product, and
        # reduce the iterator if it's partially outside of the order
        for i in range(len(prod_arg)):
            iterator = prod_arg[i]
            order = non_const_dims[i]

            # If the order is finite, we might be able to reduce the
            # cartesian product by a significant amount
            # The code below does this
            if order != 0:
                old_start, old_stop = iterator.start, iterator.stop
                prod_arg[i] = range(max(-order // 2 + 1, old_start),
                                    min(order // 2 + 1, old_stop))

        # Go through the Cartesian product / hypercube
        for prod in itertools.product(*prod_arg):

            # The first and last part of the element
            first = mod(prod[:wallnum], non_const_dims[:wallnum])
            last = mod(prod[wallnum:], non_const_dims[wallnum:])

            middle1, middle2 = mod(maxnorm_value, dim), mod(-maxnorm_value, dim)
            yield first + (middle1,) + last
            if middle1 != middle2:
                yield first + (middle2,) + last

def elements_of_maxnorm(free_rank, maxnorm_value):
    """
    Yield every element of Z^r such that max_norm(element) = maxnorm_value.

    Parameters
    ----------
    free_rank : int
        The free rank (like dimension) of Z^r, i.e. free_rank = r.
    maxnorm_value : int
        The value of the maximum norm of the elements generated.

    Yields
    -------
    tuple
        Elements in Z^r that satisfy the norm criterion.

    Examples
    ---------
    >>> free_rank = 3 # Like dimension
    >>> maxnorm_value = 4
    >>> elements = list(elements_of_maxnorm(free_rank, maxnorm_value))
    >>> # Verify that the max norm is the correct value
    >>> all(max(abs(k) for k in e) for e in elements)
    True
    >>> # Verify the number of elements
    >>> n = maxnorm_value
    >>> len(elements) == ((2*n + 1)**free_rank - (2*n - 1)**free_rank)
    True
    """
    if maxnorm_value == 0:
        yield tuple([0] * free_rank)
        return

    # There are two 'walls' per dimension, front and back
    for wall in range(free_rank):

        # In each wall, the boundaries must shrink, two at a time
        boundary_reduced = [1] * wall + [0] * (free_rank - wall - 1)

        # The arguments into the cartesian product
        prod_arg = [range(-maxnorm_value + k, maxnorm_value + 1 - k)\
                    for k in boundary_reduced]

        # Take cartesian products along the boundaries of the r-dimensional
        # cube. Yield from opposite sides of the hypercube
        for boundary_element in itertools.product(*prod_arg):
            start, end = boundary_element[:wall], boundary_element[wall:]
            yield start + (maxnorm_value,) + end
            yield start + (-maxnorm_value,) + end


def free_kernel(A):
    """
    Computes the free-to-free kernel monomorphism of A.

    Let :math:`A: \mathbb{Z}^n -> \mathbb{Z}^m` be a homomorphism from
    a free (infinite order) finitely generated Abelian group (FGA) to another
    free FGA. Associated with this homomorphism is the kernel monomorphism.
    The kernel monomorphism has the property that
    :math:`A \circ \operatorname{ker}(A) = \mathbf{0}`, where :math:`\mathbf{0}`
    denotes the zero morphism.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy integer matrix.

    Returns
    -------
    ker_A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The kernel monomorphism associated with `A`.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[1, 0, 1],
    ...             [0, 1, 1]])
    >>> ker_A = free_kernel(A)
    >>> # Verify the factorization
    >>> A * ker_A == Matrix([0, 0])
    True
    """
    U, S, V = smith_normal_form(A)
    r = len(nonzero_diag_as_list(S))
    return V[:, r:]


def free_cokernel(A):
    """
    Computes the free-to-free cokernel epimorphism of A.

    Let :math:`A: \mathbb{Z}^n -> \mathbb{Z}^m` be a homomorphism from
    a free (infinite order) finitely generated Abelian group (FGA) to another
    free FGA. Associated with this homomorphism is the cokernel epimorphism.
    The cokernel epimorphism has the property that
    :math:`\operatorname{coker}(A) \circ A = \mathbf{0}`, where
    :math:`\mathbf{0}` denotes the zero morphism.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy integer matrix.

    Returns
    -------
    coker_A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The cokernel epimorphism associated with `A`.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> from abelian.linalg.utils import matrix_mod_vector
    >>> A = Matrix([[1, 0],
    ...             [0, 1],
    ...             [1, 1]])
    >>> coker_A = free_cokernel(A)
    >>> quotient = free_quotient(A)
    >>> # Compute coker(A) * A and verify that it's 0 in the
    >>> # target group of coker(A).
    >>> product = matrix_mod_vector(coker_A * A, quotient)
    >>> product == 0 * product
    True
    """
    U, S, V = smith_normal_form(A, compute_unimod=True)
    return U


def free_image(A):
    """
    Computes the free-to-free image monomorphism of A.

    Let :math:`A: \mathbb{Z}^n -> \mathbb{Z}^m` be a homomorphism from
    a free (infinite order) finitely generated Abelian group (FGA) to another
    free FGA. Associated with this homomorphism is the image monomorphism.
    The image monomorphism has the property that :math:`A` factors through
    the composition of the coimage and image morphisms, i.e.
    :math:`\operatorname{im}(A) \circ \operatorname{coim}(A) = A`.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy integer matrix.

    Returns
    -------
    im_A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The image monomorphism associated with `A`.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[1, 0, 1],
    ...             [0, 1, 1]])
    >>> # Clearly the image is the identity matrix
    >>> free_image(A) == Matrix.eye(2)
    True
    >>> # Verify the image(A) * coimage(A) = A factorization
    >>> free_image(A) * free_coimage(A) == A
    True
    """
    U, S, V = smith_normal_form(A)
    r = len(nonzero_diag_as_list(S))
    return U.inv()[:, :r] * S[:r, :r]


def free_coimage(A):
    """
    Computes the free-to-free coimage epimorphism of A.

    Let :math:`A: \mathbb{Z}^n -> \mathbb{Z}^m` be a homomorphism from
    a free (infinite order) finitely generated Abelian group (FGA) to another
    free FGA. Associated with this homomorphism is the coimage epimorphism.
    The coimage epimorphism has the property that :math:`A` factors through
    the composition of the coimage and image morphisms, i.e.
    :math:`\operatorname{im}(A) \circ \operatorname{coim}(A) = A`.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy integer matrix.

    Returns
    -------
    coim_A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The coimage epimorphism associated with A.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[1, 0],
    ...             [0, 1],
    ...             [1, 1]])
    >>> # Clearly the image is A itself, so coim(A) must be I
    >>> free_coimage(A) == Matrix.eye(2)
    True
    >>> # Verify the image(A) * coimage(A) = A factorization
    >>> free_image(A) * free_coimage(A) == A
    True
    """
    U, S, V = smith_normal_form(A)
    r = len(nonzero_diag_as_list(S))
    return V.inv()[:r, :]


def free_quotient(A):
    """
    Compute the quotient group Z^m / im(A).

    Let :math:`A: \mathbb{Z}^n -> \mathbb{Z}^m` be a homomorphism from
    a free (infinite order) finitely generated Abelian group (FGA) to another
    free FGA. Associated with this homomorphism is the cokernel epimorphism,
    which maps from :math:`A: \mathbb{Z}^n` to
    :math:`A: \mathbb{Z}^m / \operatorname{im}(A)`.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy integer matrix.

    Returns
    -------
    quotient : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The structure of the quotient group target(A)/im(A).

    Examples
    ---------
    >>> from sympy import diag
    >>> A = diag(1, 2, 3)
    >>> free_quotient(A) == Matrix([1, 1, 6])
    True
    """
    U, S, V = smith_normal_form(A)
    m, n = A.shape
    diagonal = nonzero_diag_as_list(S)
    r = len(diagonal)
    quotient = diagonal + [0] * (m - r)
    return Matrix(quotient)


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose = False)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
from random import randint as ri
from random import shuffle
from abelian.groups import LCA
from functools import reduce
from operator import add

def random_zero_heavy(low, high):
    """
    Draw a random number, with approx 50% probability of zero.
    """
    return random.choice(list(range(low, high)) + [0]*(high - low))

def random_from_list(number, list_to_take_from):
    """
    Draw several random values from the same list.
    """
    return [random.choice(list_to_take_from) for i in range(number)]

def random_LCA(length):
    """
    Create a random LCA of a given length.
    """
    discrete = random_from_list(length, [True, False])
    periods = []
    for d in discrete:
        if d:
            # Discrete group, any integer
            periods.append(random_zero_heavy(0, 99))
        else:
            periods.append(random.choice([0, 1]))

    return LCA(orders= periods, discrete = discrete)


class TestLCA:

    @classmethod
    def setup_class(cls):
        cls.G = random_LCA(10)
        cls.H = random_LCA(10)

    def test_border_cases(self):
        """
        Border cases when initializing.
        """
        Id = LCA.trivial()

        assert Id.isomorphic(LCA([]))

    def test_canonical(self):
        """
        Test that canonical is invariant under shuffling.
        """
        G = self.G

        G_split = [grp for grp in G]
        shuffle(G_split)
        G_shuffled = reduce(add, G_split)

        assert G.canonical() == G_shuffled.canonical()

    def test_trivial_group(self):
        """
        Test the property of the trivial group.
        """
        Id = LCA.trivial()
        G = self.G
        H = self.H
        assert (G + Id).isomorphic(G)
        assert (Id + G).isomorphic(G)
        assert (H + Id).isomorphic(H)
        assert (Id + H).isomorphic(H)
        assert (Id + Id).isomorphic(Id)


    def test_rank(self):
        """
        Test the rank of LCAs.
        """
        G = self.G
        H = self.H
        assert (G + H).rank() == (G.rank() + H.rank())

    def test_length(self):
        """
        Test the length of LCAs.
        """
        G = random_LCA(ri(1, 9))
        H = random_LCA(ri(1, 9))
        assert (G + H).length() == (G.length() + H.length())

    def test_remove_trivial(self):
        """
        Test the removal of trivial subgroups.
        """
        G = self.G
        H = self.H
        after = (G + H).remove_trivial()
        before = (G.remove_trivial() + H.remove_trivial())
        assert before == after

    def test_proper_subgroup(self):
        """
        Test proper subgroups.
        """
        G = self.G
        H = G[3:-3]
        assert (H in G)

    def test_subgroup(self):
        """
        Test subgroups.
        """
        G = self.G
        assert (G in G)



if __name__ == '__main__':
    from abelian import LCA, HomLCA, LCAFunc, voronoi
    from math import exp, pi, sqrt
    Z = LCA(orders = [0], discrete = [True])
    R = LCA(orders = [0], discrete = [False])

    # Create the Gaussian function on R^2
    function = LCAFunc(lambda x: exp(-pi*sum(j**2 for j in x)), domain = R**2)

    # Create an hexagonal sampling homomorphism (lattice on R^2)
    phi = HomLCA([[1, 1/2], [0, sqrt(3)/2]], source = Z**2, target = R**2)
    phi = phi * (1/7) # Downcale the hexagon
    function_sampled = function.pullback(phi)

    # Approximate the two dimensional integral of the Gaussian
    scaling_factor = phi.A.det()
    integral_sum = 0
    for element in phi.source.elements_by_maxnorm(list(range(20))):
        integral_sum += function_sampled(element)
    print(integral_sum * scaling_factor) # 0.999999997457763

    # Sample, periodize and take DFT of the Gaussian
    phi_p = HomLCA([[10, 0], [0, 10]], source = Z**2, target = Z**2)
    periodized = function_sampled.pushforward(phi_p.cokernel())
    dual_func = periodized.dft()

    # Interpret the output of the DFT on R^2
    phi_periodize_ann = phi_p.annihilator()

    # Compute a Voronoi transversal function, interpret on R**2
    sigma = voronoi(phi.dual(), norm_p=2)
    factor = phi_p.A.det() * scaling_factor
    total_error = 0
    for element in dual_func.domain.elements_by_maxnorm():
        value = dual_func(element)
        coords_on_R = sigma(phi_periodize_ann(element))

        # The Gaussian is invariant under Fourier transformation, so we can
        # compare the error using the analytical expression
        true_val = function(coords_on_R)
        approximated_val = abs(value)
        total_error += abs(true_val - approximated_val*factor)

    assert total_error < 10e-15

